#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""Generic Tango Pool Device base classes"""

__all__ = ["PoolDevice", "PoolDeviceClass",
           "PoolElementDevice", "PoolElementDeviceClass",
           "PoolGroupDevice", "PoolGroupDeviceClass"]

__docformat__ = 'restructuredtext'

import time

from PyTango import Util, DevVoid, DevLong64, DevBoolean, DevString, \
    DevVarStringArray, DispLevel, DevState, SCALAR, SPECTRUM, \
    IMAGE, READ_WRITE, READ, AttrData, CmdArgType, DevFailed, seqStr_2_obj

from taurus.core.util.containers import CaselessDict

from sardana import InvalidId, InvalidAxis, ElementType
from sardana.pool.poolmetacontroller import DataInfo
from sardana.tango.core.SardanaDevice import SardanaDevice, SardanaDeviceClass
from sardana.tango.core.util import GenericScalarAttr, GenericSpectrumAttr, \
    GenericImageAttr, to_tango_attr_info


class PoolDevice(SardanaDevice):
    """Base Tango Pool device class"""

    #: list of extreme error states
    ExtremeErrorStates = DevState.FAULT, DevState.UNKNOWN

    #: list of busy states
    BusyStates = DevState.MOVING, DevState.RUNNING

    #: Maximum number of retries in a busy state
    BusyRetries = 3

    def __init__(self, dclass, name):
        """Constructor"""
        SardanaDevice.__init__(self, dclass, name)

    def init(self, name):
        """initialize the device once in the object lifetime. Override when
        necessary but **always** call the method from your super class
        
        :param str name: device name"""
        SardanaDevice.init(self, name)
        util = Util.instance()
        self._pool_device = util.get_device_list_by_class("Pool")[0]
        self._element = None

    @property
    def pool_device(self):
        """The tango pool device"""
        return self._pool_device

    @property
    def pool(self):
        """The sardana pool object"""
        return self.pool_device.pool

    def get_element(self):
        """Returns the underlying pool element object
        
        :return: the underlying pool element object
        :rtype: :class:`~sardana.pool.poolelement.PoolElement`"""
        return self._element

    def set_element(self, element):
        """Associates this device with the sardana element
        
        :param element: the sardana element
        :type element: :class:`~sardana.pool.poolelement.PoolElement`"""
        self._element = element

    element = property(get_element, set_element, doc="The underlying sardana element")

    def init_device(self):
        """Initialize the device. Called during startup after :meth:`init` and
        every time the tango ``Init`` command is executed.
        Override when necessary but **always** call the method from your super
        class"""
        SardanaDevice.init_device(self)

    def delete_device(self):
        """Clean the device. Called during shutdown and every time the tango
        ``Init`` command is executed.
        Override when necessary but **always** call the method from your super
        class"""
        SardanaDevice.delete_device(self)

    def Abort(self):
        """The tango abort command. Aborts the active operation"""
        self.element.abort()
        try:
            self.element.get_state(cache=False, propagate=2)
        except:
            self.warning("Abort: failed to read state")

    def is_Abort_allowed(self):
        """Returns True if it is allowed to execute the tango abort command
        
        :return: True if it is allowed to execute the tango abort command or
                 False otherwise
        :rtype: bool"""
        return self.get_state() != DevState.UNKNOWN

    def Stop(self):
        """The tango stop command. Stops the active operation"""
        self.element.stop()
        try:
            self.element.get_state(cache=False, propagate=2)
        except:
            self.info("Stop: failed to read state")

    def is_Stop_allowed(self):
        """Returns True if it is allowed to execute the tango stop command
        
        :return: True if it is allowed to execute the tango stop command or
                 False otherwise
        :rtype: bool"""
        return self.get_state() != DevState.UNKNOWN

    def _is_allowed(self, req_type):
        """Generic is_allowed"""
#        state = self.get_state()
#        if state in self.ExtremeErrorStates:
#            return False
#        if req_type == AttReqType.WRITE_REQ:
#            if state in self.BusyStates:
#                return False
        return True

    def get_dynamic_attributes(self):
        """Returns the standard dynamic and fully dynamic attributes for this
        device. The return is a tuple of two dictionaries:
        
        - standard attributes: caseless dictionary with key being the attribute
          name and value is a tuple of attribute name(str), tango information,
          attribute information
        - dynamic attributes: caseless dictionary with key being the attribute
          name and value is a tuple of attribute name(str), tango information,
          attribute information
        
        **tango information**
            seq< :class:`~PyTango.CmdArgType`, :class:`~PyTango.AttrDataFormat`, :class:`~PyTango.AttrWriteType` >
        
        **attribute information**
            attribute information as returned by the sardana controller
        
        :return: the standard dynamic and fully dynamic attributes
        :rtype: seq< :class:`~taurus.core.util.CaselessDict`, :class:`~taurus.core.util.CaselessDict`\>
        """
        return CaselessDict(), CaselessDict()

    def initialize_dynamic_attributes(self):
        """Initializes this device dynamic attributes"""
        self._attributes = attrs = CaselessDict()

        attr_data = self.get_dynamic_attributes()

        std_attrs, dyn_attrs = attr_data
        self.remove_unwanted_dynamic_attributes(std_attrs, dyn_attrs)

        if std_attrs is not None:
            read = self.__class__._read_DynamicAttribute
            write = self.__class__._write_DynamicAttribute
            is_allowed = self.__class__._is_DynamicAttribute_allowed
            for attr_name, data_info in std_attrs.items():
                attr_name, data_info, attr_info = data_info
                attr = self.add_standard_attribute(attr_name, data_info,
                                                   attr_info, read,
                                                   write, is_allowed)
                attrs[attr.get_name()] = None

        if dyn_attrs is not None:
            read = self.__class__._read_DynamicAttribute
            write = self.__class__._write_DynamicAttribute
            is_allowed = self.__class__._is_DynamicAttribute_allowed
            for attr_name, data_info in dyn_attrs.items():
                attr_name, data_info, attr_info = data_info
                attr = self.add_dynamic_attribute(attr_name, data_info,
                                                  attr_info, read,
                                                  write, is_allowed)
                attrs[attr.get_name()] = None
        return attrs

    def remove_unwanted_dynamic_attributes(self, new_std_attrs, new_dyn_attrs):
        """Removes unwanted dynamic attributes from previous device creation"""

        dev_class = self.get_device_class()
        multi_attr = self.get_device_attr()
        multi_class_attr = dev_class.get_class_attr()
        static_attr_names = map(str.lower, dev_class.attr_list.keys())
        static_attr_names.extend(('state', 'status'))

        new_attrs = CaselessDict(new_std_attrs)
        new_attrs.update(new_dyn_attrs)

        device_attr_names = []
        for i in range(multi_attr.get_attr_nb()):
            device_attr_names.append(multi_attr.get_attr_by_ind(i).get_name())

        for attr_name in device_attr_names:
            attr_name_lower = attr_name.lower()
            if attr_name_lower in static_attr_names:
                continue
            try:
                self.remove_attribute(attr_name)
            except:
                self.warning("Error removing dynamic attribute %s",
                             attr_name_lower)
                self.debug("Details:", exc_info=1)

        klass_attr_names = []
        klass_attrs = multi_class_attr.get_attr_list()
        for ind in range(len(klass_attrs)):
            klass_attr_names.append(klass_attrs[ind].get_name())

        for attr_name in klass_attr_names:
            attr_name_lower = attr_name.lower()
            if attr_name_lower in static_attr_names:
                continue
            # if new dynamic attribute is in class attribute then delete it
            # from class attribute to be later on added again (eventually
            # with diffent data type or data format)
            if attr_name_lower in new_attrs:
                try:
                    attr = multi_class_attr.get_attr(attr_name)

                    old_type = CmdArgType(attr.get_type())
                    old_format = attr.get_format()
                    old_access = attr.get_writable()

                    new_attr = new_attrs[attr_name]
                    new_type, new_format, new_access = new_attr[1][0][:3]
                    differ = new_type != old_type or \
                             new_format != old_format or \
                             new_access != old_access
                    if differ:
                        self.info("Replacing dynamic attribute %s", attr_name)
                        self.debug("old type: %s, new type: %s",
                                   old_type, new_type)
                        self.debug("old format: %s, new format: %s",
                                   old_format, new_format)
                        self.debug("old access: %s, new access: %s",
                                   old_access, new_access)
                        multi_class_attr.remove_attr(attr.get_name(),
                                                     attr.get_cl_name())
                except:
                    self.warning("Error removing dynamic attribute %s from "\
                                 " device class", attr_name)
                    self.debug("Details:", exc_info=1)

    def add_dynamic_attribute(self, attr_name, data_info, attr_info, read,
                              write, is_allowed):
        """Adds a single dynamic attribute
        
        :param str attr_name: the attribute name
        :param data_info: tango attribute information
        :type data_info: seq< :class:`~PyTango.CmdArgType`, :class:`~PyTango.AttrDataFormat`, :class:`~PyTango.AttrWriteType` >
        :param attr_info: attribute information
        :param read: read method for the attribute
        :param write: write method for the attribute
        :param is_allowed: is allowed method"""
        tg_type, tg_format, tg_access = data_info[0]

        if tg_access == READ:
            write = None
        if tg_format == SCALAR:
            attr = GenericScalarAttr(attr_name, tg_type, tg_access)
        if tg_format == SPECTRUM:
            dim_x = attr_info.maxdimsize[0]
            attr = GenericSpectrumAttr(attr_name, tg_type, tg_access,
                                       dim_x=dim_x)
        elif tg_format == IMAGE:
            dim_x, dim_y = attr_info.maxdimsize
            attr = GenericImageAttr(attr_name, tg_type, tg_access,
                                    dim_x=dim_x, dim_y=dim_y)

        if tg_access == READ_WRITE and tg_format == SCALAR:
            memorized = attr_info.memorized.lower()
            if memorized == 'true':
                attr.set_memorized()
                attr.set_memorized_init(True)
            elif memorized == 'true_without_hard_applied':
                attr.set_memorized()
                attr.set_memorized_init(False)
        attr.set_disp_level(DispLevel.EXPERT)
        return self.add_attribute(attr, read, write, is_allowed)

    def add_standard_attribute(self, attr_name, data_info, attr_info, read,
                               write, is_allowed):
        """Adds a single standard dynamic attribute
        
        :param str attr_name: the attribute name
        :param data_info: tango attribute information
        :type data_info: seq< :class:`~PyTango.CmdArgType`, :class:`~PyTango.AttrDataFormat`, :class:`~PyTango.AttrWriteType` >
        :param attr_info: attribute information
        :param read: read method for the attribute
        :param write: write method for the attribute
        :param is_allowed: is allowed method"""
        dev_class = self.get_device_class()
        attr_data = AttrData(attr_name, dev_class.get_name(), data_info)
        attr = self.add_attribute(attr_data, read, write, is_allowed)
        return attr

    def read_DynamicAttribute(self, attr):
        """Generic read dynamic attribute.
        Default implementation raises :exc:`NotImplementedError`
        
        :param attr: attribute to be read
        :type attr: :class:`~PyTango.Attribute`
        
        :raises: :exc:`NotImplementedError`"""
        raise NotImplementedError

    def write_DynamicAttribute(self, attr):
        """Generic write dynamic attribute.
        Default implementation raises :exc:`NotImplementedError`
        
        :param attr: attribute to be written
        :type attr: :class:`~PyTango.Attribute`
        
        :raises: :exc:`NotImplementedError`"""
        raise NotImplementedError

    def is_DynamicAttribute_allowed(self, req_type):
        """Generic is dynamic attribute allowed.
        Default implementation calls :meth:`_is_allowed`
        
        :param req_type: request type
        :type attr: :class:`~PyTango.AttrRequestType`"""
        return self._is_allowed(req_type)

    def _read_DynamicAttribute(self, attr):
        """Generic internal read dynamic attribute.
        Checks if this object has a 'read_'+<attr_name> method and calls it.
        If not calls :meth:`read_DynamicAttribute`.
        
        :param attr: attribute to be read
        :type attr: :class:`~PyTango.Attribute`"""
        name = attr.get_name()

        read_name = "read_" + name
        if hasattr(self, read_name):
            read = getattr(self, read_name)
            return read(attr)

        return self.read_DynamicAttribute(attr)

    def _write_DynamicAttribute(self, attr):
        """Generic internal write dynamic attribute.
        Checks if this object has a 'write_'+<attr_name> method and calls it.
        If not calls :meth:`write_DynamicAttribute`.
        
        :param attr: attribute to be written
        :type attr: :class:`~PyTango.Attribute`"""
        name = attr.get_name()
        write_name = "write_" + name
        if hasattr(self, write_name):
            write = getattr(self, write_name)
            return write(attr)
        return self.write_DynamicAttribute(attr)

    def _is_DynamicAttribute_allowed(self, req_type):
        """Generic is dynamic attribute allowed.
        Default implementation calls :meth:`is_DynamicAttribute_allowed`
        
        :param req_type: request type
        :type attr: :class:`~PyTango.AttrRequestType`"""
        return self.is_DynamicAttribute_allowed(req_type)

    def dev_state(self):
        """Calculates and returns the device state. Called by Tango on a read
        state request.
        
        :return: the device state
        :rtype: :class:`~PyTango.DevState`"""
        element = self.element
        try:
            use_cache = element.is_in_operation() and not self.Force_HW_Read
            ctrl_state = element.get_state(cache=use_cache, propagate=0)
            state = self.calculate_tango_state(ctrl_state)
            return state
        except:
            self.error("Exception trying to return state")
            self.debug("Details:", exc_info=1)
            return DevState.FAULT

    def dev_status(self):
        """Calculates and returns the device status. Called by Tango on a read
        status request.
        
        :return: the device status
        :rtype: str"""
        element = self.element
        try:
            use_cache = element.is_in_operation() and not self.Force_HW_Read
            ctrl_status = self.element.get_status(cache=use_cache, propagate=0)
            status = self.calculate_tango_status(ctrl_status)
            return status
        except Exception, e:
            msg = "Exception trying to return status: %s" % str(e)
            self.error(msg)
            self.debug("Details:", exc_info=1)
            return msg

    def wait_for_operation(self):
        """Waits for an operation to finish. It uses the maxumum number of
        retries. Sleeps 0.01s between retries.
        
        :raises: :exc:`Exception` in case of a timeout"""
        element, n = self.element, self.BusyRetries
        while element.is_in_operation():
            if n == 0:
                raise Exception("Wait for operation timedout")
            time.sleep(0.01)
            self.warning("waited for operation")
            n = n - 1

    def Restore(self):
        """Restore tango command. Restores the attributes to their former glory.
        This applies to memorized writable attributes which have a set point
        stored in the database"""
        restore_attributes, db_values = self.get_restore_data()
        multi_attribute = self.get_device_attr()

        for attr_name in restore_attributes:
            props = db_values[attr_name]
            if props is None or not "__value" in props:
                continue
            attribute = multi_attribute.get_w_attr_by_name(attr_name)
            write_meth_name = "write_" + attr_name
            write_meth = getattr(self, write_meth_name, None)
            if write_meth is None:
                self.warning("Could not recover %s: %s does not exist",
                             attr_name, write_meth_name)
                continue
            self.restore_attribute(attribute, write_meth, props['__value'])

    def get_restore_data(self):
        restore_attributes = self.get_attributes_to_restore()
        db = Util.instance().get_database()
        db_values = db.get_device_attribute_property(self.get_name(),
                                                     restore_attributes)
        return restore_attributes, db_values

    def get_attributes_to_restore(self):
        std_attrs, dyn_attrs = self.get_dynamic_attributes()
        multi_attribute = self.get_device_attr()

        restore = []
        for attr_name in std_attrs:
            try:
                attribute = multi_attribute.get_w_attr_by_name(attr_name)
            except DevFailed:
                continue
            restore.append(attribute.get_name())
        for attr_name in dyn_attrs:
            try:
                attribute = multi_attribute.get_w_attr_by_name(attr_name)
            except DevFailed:
                continue
            restore.append(attribute.get_name())
        return restore

    def _get_attribute_value_from_db_value(self, attribute, db_value):
        value = seqStr_2_obj(db_value, attribute.get_data_type(),
                             attribute.get_data_format())
        return value

    def restore_attribute(self, attribute, write_meth, db_value):
        value = self._get_attribute_value_from_db_value(attribute, db_value)
        attr_name = attribute.get_name()
        try:
            attribute.set_write_value(value)
            self.info("Restoring %s", attr_name)
            write_meth(attribute)
        except:
            self.warning("Could not recover %s: Error in write", attr_name)
            self.debug("Details:", exc_info=1)



class PoolDeviceClass(SardanaDeviceClass):
    """Base Tango Pool Device Class class"""

    #:
    #: Sardana device class properties definition
    #:
    #: .. seealso:: :ref:`server`
    #:
    class_property_list = SardanaDeviceClass.class_property_list

    #:
    #: Sardana device properties definition
    #:
    #: .. seealso:: :ref:`server`
    #:
    device_property_list = {
        'Id'            : [DevLong64, "Internal ID", InvalidId ],
        'Force_HW_Read' : [DevBoolean, "Force a hardware read of value even "
                                       "when in operation (motion/acquisition",
                           False],
    }
    device_property_list.update(SardanaDeviceClass.device_property_list)

    #:
    #: Sardana device command definition
    #:
    #: .. seealso:: :ref:`server`
    #:
    cmd_list = {
        'Stop'    : [ [DevVoid, ""], [DevVoid, ""] ],
        'Abort'   : [ [DevVoid, ""], [DevVoid, ""] ],
        'Restore' : [ [DevVoid, ""], [DevVoid, ""] ],
    }
    cmd_list.update(SardanaDeviceClass.cmd_list)

    #:
    #: Sardana device attribute definition
    #:
    #: .. seealso:: :ref:`server`
    #:
    attr_list = {
    }
    attr_list.update(SardanaDeviceClass.attr_list)

    standard_attr_list = {}

    def _get_class_properties(self):
        ret = SardanaDeviceClass._get_class_properties(self)
        ret['Description'] = "Generic Pool device class"
        ret['InheritedFrom'].insert(0, 'SardanaDevice')
        return ret


class PoolElementDevice(PoolDevice):
    """Base Tango Pool Element Device class"""

    def init_device(self):
        """Initialize the device. Called during startup after :meth:`init` and
        every time the tango ``Init`` command is executed.
        Override when necessary but **always** call the method from your super
        class"""
        PoolDevice.init_device(self)

        self.instrument = None
        self.ctrl = None
        try:
            instrument_id = int(self.Instrument_id)
            if instrument_id != InvalidId:
                instrument = self.pool.get_element_by_id(instrument_id)
                self.instrument = instrument
        except ValueError:
            pass
        try:
            ctrl_id = int(self.Ctrl_id)
            if ctrl_id != InvalidId:
                ctrl = self.pool.get_element_by_id(ctrl_id)
                self.ctrl = ctrl
        except ValueError:
            pass

    def read_Instrument(self, attr):
        """Read the value of the ``Instrument`` tango attribute.
        Returns the instrument full name or empty string if this element doesn't
        belong to any instrument
        
        :param attr: tango instrument attribute
        :type attr: :class:`~PyTango.Attribute`"""
        instrument = self.element.instrument
        if instrument is None:
            attr.set_value('')
        else:
            attr.set_value(instrument.full_name)

    def write_Instrument(self, attr):
        """Write the value of the ``Instrument`` tango attribute.
        Sets a new instrument full name or empty string if this element doesn't
        belong to any instrument.
        The instrument **must** have been previously created.
        
        :param attr: tango instrument attribute
        :type attr: :class:`~PyTango.Attribute`"""
        name = attr.get_write_value()
        instrument = None
        if name:
            instrument = self.pool.get_element(full_name=name)
            if instrument.get_type() != ElementType.Instrument:
                raise Exception("%s is not an instrument" % name)
        self.element.instrument = instrument
        db = Util.instance().get_database()
        db.put_device_property(self.get_name(), { "Instrument_id" : instrument.id })

    def get_dynamic_attributes(self):
        """Override of :class:`PoolDevice.get_dynamic_attributes`.
        Returns the standard dynamic and fully dynamic attributes for this
        device. The return is a tuple of two dictionaries:
        
        - standard attributes: caseless dictionary with key being the attribute
          name and value is a tuple of attribute name(str), tango information,
          attribute information
        - dynamic attributes: caseless dictionary with key being the attribute
          name and value is a tuple of attribute name(str), tango information,
          attribute information
        
        **tango information**
            seq< :class:`~PyTango.CmdArgType`, :class:`~PyTango.AttrDataFormat`, :class:`~PyTango.AttrWriteType` >
        
        **attribute information**
            attribute information as returned by the sardana controller
        
        :return: the standard dynamic and fully dynamic attributes
        :rtype: seq< :class:`~taurus.core.util.CaselessDict`, :class:`~taurus.core.util.CaselessDict`\>
        """

        if hasattr(self, "_dynamic_attributes_cache"):
            return self._standard_attributes_cache, self._dynamic_attributes_cache
        ctrl = self.ctrl
        if ctrl is None:
            self.warning("no controller: dynamic attributes NOT created")
            return PoolDevice.get_dynamic_attributes(self)
        if not ctrl.is_online():
            self.warning("controller offline: dynamic attributes NOT created")
            return PoolDevice.get_dynamic_attributes(self)

        self._dynamic_attributes_cache = dyn_attrs = CaselessDict()
        self._standard_attributes_cache = std_attrs = CaselessDict()
        dev_class = self.get_device_class()
        axis_attrs = ctrl.get_axis_attributes(self.element.axis)

        std_attrs_lower = [ attr.lower() for attr in dev_class.standard_attr_list ]
        for attr_name, attr_info in axis_attrs.items():
            attr_name_lower = attr_name.lower()
            if attr_name_lower in std_attrs_lower:
                data_info = DataInfo.toDataInfo(attr_name, attr_info)
                tg_info = dev_class.standard_attr_list[attr_name]
                std_attrs[attr_name] = attr_name, tg_info, data_info
            else:
                data_info = DataInfo.toDataInfo(attr_name, attr_info)
                name, tg_info = to_tango_attr_info(attr_name, data_info)
                dyn_attrs[attr_name] = name, tg_info, data_info
        return std_attrs, dyn_attrs

    def read_DynamicAttribute(self, attr):
        """Read a generic dynamic attribute. Calls the controller of this
        element to get the dynamic attribute value
        
        :param attr: tango attribute
        :type attr: :class:`~PyTango.Attribute`"""
        name = attr.get_name()
        ctrl = self.ctrl
        if ctrl is None:
            raise Exception("Cannot read %s. Controller not build!" % name)
        v = ctrl.get_axis_attr(self.element.axis, name)
        if v is None:
            raise TypeError("Cannot read %s. Controller returns None" % (name,))
        attr.set_value(v)

    def write_DynamicAttribute(self, attr):
        """Write a generic dynamic attribute. Calls the controller of this
        element to get the dynamic attribute value
        
        :param attr: tango attribute
        :type attr: :class:`~PyTango.Attribute`"""
        name = attr.get_name()
        value = attr.get_write_value()
        self.debug("writing dynamic attribute %s with value %s", name, value)
        ctrl = self.ctrl
        if ctrl is None:
            raise Exception("Cannot write %s. Controller not build!" % name)
        ctrl.set_axis_attr(self.element.axis, name, value)

    def read_SimulationMode(self, attr):
        """Read the current simulation mode. 
        
        :param attr: tango attribute
        :type attr: :class:`~PyTango.Attribute`"""
        attr.set_value(self.element.simulation_mode)

    def write_SimulationMode(self, attr):
        """Sets the simulation mode. 
        
        :param attr: tango attribute
        :type attr: :class:`~PyTango.Attribute`"""
        self.element.simulation_mode = attr.get_write_value()


class PoolElementDeviceClass(PoolDeviceClass):
    """Base Tango Pool Element Device Class class"""

    #:
    #: Sardana device properties definition
    #:
    #: .. seealso:: :ref:`server`
    #:
    device_property_list = {
        "Axis"          : [ DevLong64, "Axis in the controller", [ InvalidAxis ] ],
        "Ctrl_id"       : [ DevLong64, "Controller ID", [ InvalidId ] ],
        "Instrument_id" : [ DevLong64, "Controller ID", [ InvalidId ] ],
    }
    device_property_list.update(PoolDeviceClass.device_property_list)

    #:
    #: Sardana device attribute definition
    #:
    #: .. seealso:: :ref:`server`
    #:
    attr_list = {
        'Instrument' :    [ [DevString, SCALAR, READ_WRITE],
                          { 'label'         : "Instrument",
                            'Display level' : DispLevel.EXPERT } ],
        'SimulationMode': [ [DevBoolean, SCALAR, READ_WRITE],
                          { 'label'         : "Simulation mode" } ],
    }
    attr_list.update(PoolDeviceClass.attr_list)

    cmd_list = {
    }
    cmd_list.update(PoolDeviceClass.cmd_list)

    def get_standard_attr_info(self, attr):
        """Returns information about the standard attribute
        
        :param str attr: attribute name
        :return: a sequence of tango data_type, data format"""
        return self.standard_attr_list[attr]

    def _get_class_properties(self):
        ret = PoolDeviceClass._get_class_properties(self)
        ret['Description'] = "Generic Pool element device class"
        ret['InheritedFrom'].insert(0, 'PoolDevice')
        return ret


class PoolGroupDevice(PoolDevice):
    """Base Tango Pool Group Device class"""

    def read_ElementList(self, attr):
        """Read the element list. 
        
        :param attr: tango attribute
        :type attr: :class:`~PyTango.Attribute`"""
        attr.set_value(self.get_element_names())

    def get_element_names(self):
        """Returns the list of element names. 
        
        :return: a list of attribute names"""
        elements = self.element.get_user_elements()
        return [ element.name for element in elements ]

    def elements_changed(self, evt_src, evt_type, evt_value):
        """Callback for when the elements of this group changed"""
        self.push_change_event("ElementList", self.get_element_names())


class PoolGroupDeviceClass(PoolDeviceClass):
    """Base Tango Pool Group Device Class class"""

    #:
    #: Sardana device properties definition
    #:
    #: .. seealso:: :ref:`server`
    #:
    device_property_list = {
        "Elements" :    [ DevVarStringArray, "elements in the group", [ ] ],
    }
    device_property_list.update(PoolDeviceClass.device_property_list)

    #:
    #: Sardana device command definition
    #:
    #: .. seealso:: :ref:`server`
    #:
    cmd_list = {
    }
    cmd_list.update(PoolDeviceClass.cmd_list)

    #:
    #: Sardana device attribute definition
    #:
    #: .. seealso:: :ref:`server`
    #:
    attr_list = {
        'ElementList'  : [ [ DevString, SPECTRUM, READ, 4096] ],
    }
    attr_list.update(PoolDeviceClass.attr_list)

    def _get_class_properties(self):
        ret = PoolDeviceClass._get_class_properties(self)
        ret['Description'] = "Generic Pool group device class"
        ret['InheritedFrom'].insert(0, 'PoolDevice')
        return ret



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module is part of the Python Sardana libray. It defines the base class
for Sardana manager"""

from __future__ import absolute_import

__all__ = ["SardanaElementManager", "SardanaIDManager"]

__docformat__ = 'restructuredtext'

from taurus.core.util.codecs import CodecFactory

from sardana import InvalidId


class SardanaElementManager(object):
    """A class capable of manage elements"""
    
    SerializationProtocol = 'json'
    
    def get_serialization_protocol(self):
        return self.SerializationProtocol
    
    def set_serialization_protocol(self, protocol):
        self.SerializationProtocol = protocol
    
    serialization_protocol = property(get_serialization_protocol,
                                      set_serialization_protocol,
                                      doc="the serialization protocol")
    
    def serialize_element(self, element, *args, **kwargs):
        obj = element.serialize(*args, **kwargs)
        return self.serialize_object(obj)
    
    def serialize_object(self, obj, *args, **kwargs):
        return CodecFactory().encode(self.serialization_protocol, ('', obj),
                                     *args, **kwargs)[1]
    
    def str_element(self, element, *args, **kwargs):
        obj = element.serialize(*args, **kwargs)
        return self.str_object(obj)
    
    def str_object(self, obj, *args, **kwargs):
        # TODO: use the active codec instead of hardcoded json
        return  CodecFactory().encode('json', ('', obj), *args, **kwargs)[1]


class SardanaIDManager(object):
    """A class capable of manage ids"""
    
    _last_id = InvalidId
    
    def get_new_id(self):
        """Returns a new ID. The ID becomes reserved at this moment.
        
        :return: a new ID
        :rtype: int"""
        self._last_id += 1
        return self._last_id
    
    def rollback_id(self):
        """Free previously reserved ID"""
        self._last_id -= 1
    
    def reserve_id(self, nid):
        """Marks the given ID as reserved
        
        :param id: the ID to be reserved
        :type id: int"""
        assert type(nid) == int
        if nid > self._last_id:
            self._last_id = nid
    


#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""Tests for scan macros"""

from taurus.external import unittest
from sardana.macroserver.macros.test import (RunStopMacroTestCase,
                                             testRun, testStop, SarDemoEnv)

#get handy motor names from sardemo
try:
    _MOTORS = SarDemoEnv().getMotors()
    _m1, _m2 = _MOTORS[:2]
except RuntimeError:
    import taurus
    from sardana import sardanacustomsettings
    door_name = getattr(sardanacustomsettings, 'UNITTEST_DOOR_NAME',
                        'UNDEFINED')
    taurus.warning("The door %s is not running. " % (door_name) +
                   "Ignore this message if you are building the documentation")
    _m1 = _m2 = 'motor_not_defined'
except Exception, e:
    import taurus
    taurus.debug(e)
    taurus.warning("It was not possible to retrieve the motor names. " +
                 "Ignore this message if you are building the documentation.")
    _m1 = _m2 = 'motor_not_defined'


def parsing_log_output(log_output):
    """A helper method to parse log output of an executed scan macro.
    :params log_output: (seq<str>) Result of macro_executor.getLog('output')
    (see description in :class:`.BaseMacroExecutor`).

    :return: (seq<number>) The numeric data of a scan.
    """
    first_data_line = 1
    scan_index = 0
    data = []
    for line, in log_output[first_data_line:]:
        # Get a list of elements without white spaces between them
        l = line.split()
        # Cast all elements of the scan line (l) to float
        l = [float(scan_elem) for scan_elem in l]
        # Cast index of scan to int (the first element of the list)
        l[scan_index] = int(l[scan_index])
        data.append(l)
    return data


class ANscanTest(RunStopMacroTestCase):

    """Not yet implemented. Once implemented it will test anscan.
    See :class:`.RunStopMacroTestCase` for requirements.
    """
    pass


class DNscanTest(ANscanTest):

    """Not yet implemented. Once implemented it will test the macro dnscanc.
    See :class:`ANscanTest` for requirements.
    """
    pass


class DNscancTest(DNscanTest):

    """Not yet implemented. Once implemented it will test the macro dnscanc.
    See :class:`DNscanTest` for requirements.
    """
    pass


@testRun(macro_params=[_m1, '0', '5', '4', '.1'], wait_timeout=float("inf"))
@testStop(macro_params=[_m1, '0', '5', '3', '.1'])
class AscanTest(ANscanTest, unittest.TestCase):

    """Test of ascan macro. See :class:`ANscanTest` for requirements.
    It verifies that macro ascan can be executed and stoped and tests
    the output of the ascan using data from log system and macro data.
    """
    macro_name = 'ascan'

    def macro_runs(self, macro_params=None, wait_timeout=float("inf")):
        """Reimplementation of macro_runs method for ascan macro.
        It verifies using double checking, with log output and data from
        the macro:

            - The motor initial and final positions of the scan are the
              ones given as input.

            - Intervals in terms of motor position between one point and
              the next one are equidistant.
        """
        #call the parent class implementation
        ANscanTest.macro_runs(self, macro_params=macro_params,
                              wait_timeout=wait_timeout)

        mot_name = macro_params[0]
        expected_init_pos = float(macro_params[1])
        expected_final_pos = float(macro_params[2])
        self.steps = int(macro_params[-2])
        interval = abs(expected_final_pos - expected_init_pos) / self.steps

        # Test data from macro (macro_executor.getData())
        data = self.macro_executor.getData()
        mot_init_pos = data[min(data.keys())].data[mot_name]
        mot_final_pos = data[max(data.keys())].data[mot_name]
        pre = mot_init_pos

        for step in range(1, max(data.keys()) + 1):
            self.assertAlmostEqual(abs(pre - data[step].data[mot_name]),
                                   interval, 7,
                                   "Step interval differs for more than expected (using getData)")
            pre = data[step].data[mot_name]

        self.assertAlmostEqual(mot_init_pos, expected_init_pos, 7,
                               "Initial possition differs from set value (using getData)")
        self.assertAlmostEqual(mot_final_pos, expected_final_pos, 7,
                               "Final possition differs from set value (using getData)")

        # Test data from log_output (macro_executor.getLog('output'))
        log_output = self.macro_executor.getLog('output')
        data = parsing_log_output(log_output)
        init_pos = 0
        last_pos = -1
        value = 1
        pre = data[init_pos]
        for step in data[1:]:
            self.assertAlmostEqual(abs(pre[value] - step[value]),
                                   interval, 7,
                                   "Step interval differs for more than expected (using getData)")
            pre = step

        self.assertAlmostEqual(data[init_pos][value], expected_init_pos, 7,
                               "Initial possition differs from set value (using getLog)")
        self.assertAlmostEqual(data[last_pos][value], expected_final_pos, 7,
                               "Final possition differs from set value (using getLog)")


@testRun(macro_params=[_m1, '-1', '1', '2', '.1'])
@testStop(macro_params=[_m1, '1', '-1', '3', '.1'])
class DscanTest(DNscanTest, unittest.TestCase):

    """Test of dscan macro. It verifies that macro dscan can be executed and
    stoped. See :class:`DNscanTest` for requirements.
    """
    macro_name = 'dscan'


@testRun(macro_params=[_m1, '-1', '1', '3', _m2, '-1', '0', '2', '.1'])
@testRun(macro_params=[_m1, '-2', '2', '3', _m2, '-2', '-1', '2', '.1'])
@testStop(macro_params=[_m1, '-3', '0', '3', _m2, '-3', '0', '2', '.1'])
class MeshTest(RunStopMacroTestCase, unittest.TestCase):

    """Test of mesh macro. It verifies that macro mesh can be executed and
    stoped. See :class:`.RunStopMacroTestCase` for requirements.
    """
    macro_name = 'mesh'



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module provides a base widget that can be used to display a taurus
model in a table widget"""

__all__ = ["MntGrpChannelEditor"]

__docformat__ = 'restructuredtext'

from taurus.core.taurusdevice import TaurusDevice
from taurus.external.qt import Qt
from taurus.qt.qtcore.model import TaurusBaseTreeItem, TaurusBaseModel
from taurus.qt.qtgui.model import EditorToolBar
from taurus.qt.qtgui.resource import getIcon, getThemeIcon
from taurus.qt.qtgui.table import TaurusBaseTableWidget
from taurus.qt.qtgui.panel import TaurusModelChooser
from taurus.core.taurusbasetypes import TaurusElementType

from sardana.taurus.core.tango.sardana import ChannelView, PlotType, \
    Normalization, AcqTriggerType
from sardana.taurus.core.tango.sardana.pool import getChannelConfigs

#===============================================================================
# some dummydict for developing the "Experimental Configuration widget"
# This block is to be removed and the dictionaries will be defined and
# initialized in Sardana's Door code


# dict <str, obj> with (at least) keys:
#    - 'timer' : the timer channel name / timer channel id
#    - 'monitor' : the monitor channel name / monitor channel id
#    - 'controllers' : dict<Controller, dict> where:
#        - key: ctrl
#        - value: dict<str, dict> with (at least) keys:
#            - 'units': dict<str, dict> with (at least) keys:
#                - 'id' : the unit ID inside the controller
#                - 'timer' : the timer channel name / timer channel id
#                - 'monitor' : the monitor channel name / monitor channel id
#                - 'trigger_type' : a value from AcqTriggerType enum
#                - 'channels' where value is a dict<str, obj> with (at least) keys:
#                    - 'index' : int indicating the position of the channel in the measurement group
#                    - 'id' : the channel name ( channel id )
#                    optional keys:
#                    - 'enabled' : True/False (default is True)
#                    any hints:
#                    - 'output' : True/False (default is True)
#                    - 'plot_type' : a value from PlotType enum
#                    - 'plot_axes' : list<str> 'where str is channel name/'step#/'index#' (default is [])
#                    - 'label' : prefered label (default is channel name)
#                    - 'scale' : <float, float> with min/max (defaults to channel
#                                range if it is defined
#                    - 'plot_color' : int representing RGB
#    optional keys:
#    - 'label' : measurement group label (defaults to measurement group name)
#    - 'description' : measurement group description


#===============================================================================
DEFAULT_STRING_LENGTH = 80  #just an arbitrary value to use as default string length...

def createChannelDict(channel, index=None, **kwargs):
    from taurus.core.tango import FROM_TANGO_TO_STR_TYPE
    import PyTango
    import numpy

    if isinstance(channel, (str, unicode)):
        #@fixme: to make things uglier, I lazily assume Tango attribute namin
        dev_name, attr_name = channel.rsplit('/', 1)
        name = attr_name
        try:
            dev = PyTango.DeviceProxy(dev_name)
            db = dev.get_device_db()
            try:
                alias = db.get_alias(dev.name())
            except:
                # no alias...
                alias = dev.name()
            label = alias + "/" + attr_name
        except:
            label = channel
        full_name = channel
        source = channel
    else:
        name = channel['name']
        label = name
        full_name = channel['full_name']
        source = channel['source']

    ret = {
           'name' : name,
           'label': label,
           'full_name': full_name,
           'enabled': True,  # bool. Whether this channel is enabled (if not enabled, it won't be used for output or plot)
           'output': True,  # bool. Whether to show output in the stdout
           'data_type':'float64',
           'data_units': 'No unit',
#           'timer': '', #should contain a channel name
#           'monitor': '', #should contain a channel name
#           'trigger': '', #should contain a channel name
           'conditioning': '',  #this is a python expresion to be evaluated for conditioning the data. The data for this channel can be referred as 'x' and data from other channels can be referred by channel name
           'normalization': Normalization.No,  # one of the Normalization enumeration members
           'nexus_path': '',  #string indicating the location of the data of this channel within the nexus tree
           }

    #If the channel is a Tango one, try to guess data_type, shape and data_units
    attrproxy = attrconf = value = None
    dtype = shape = None
    try:
        attrproxy = PyTango.AttributeProxy(source)
        attrconf = attrproxy.get_config()
        # avoid trying to read for scalars. We know that their shape must be ()
        if attrconf.data_format != PyTango.AttrDataFormat.SCALAR:
            value = attrproxy.read().value
    except Exception, e:
        print str(e)

    if value is not None:
        shape = list(numpy.shape(value))
        dtype = getattr(value, 'dtype', numpy.dtype(type(value))).name
        ret['data_units'] = attrconf.unit
    elif attrconf is not None:
        if attrconf.data_format == PyTango.AttrDataFormat.SCALAR:
            shape = []
        else:
            shape = [n for n in (attrconf.max_dim_x, attrconf.max_dim_y) if n > 0]
        dtype = FROM_TANGO_TO_STR_TYPE[attrconf.data_type]
        ret['data_units'] = attrconf.unit

    if dtype is not None:
#        if dtype.startswith('str'):
#            dtype='char'
#            shape = list(shape)+[DEFAULT_STRING_LENGTH]
#        elif dtype == 'bool':
#            dtype='int8'
        ret['data_type'] = dtype
    if shape is not None:
        ret['shape'] = shape

    #now overwrite using the arguments
    ret.update(kwargs)

    #Calculate the index
    if index is not None:
        ret['index'] = index  #an integer used for ordering the channel in this measurement group

    #Choose a default plot_type for the channel
    if 'plot_type' not in ret:
        default_plot_type = {0:PlotType.Spectrum, 2:PlotType.Image, None:PlotType.No}
        try: rank = len(ret['shape'])
        except KeyError:
            rank = None  #if shape is not known, use the default plot_type
        ret['plot_type'] = default_plot_type.get(rank, PlotType.No)

    #And a default value for plot_axes
    if 'plot_axes' not in ret:
        default_axes = {PlotType.No:[], PlotType.Spectrum:['<mov>'], PlotType.Image:['<idx>', '<idx>']}
        ret['plot_axes'] = default_axes[ret['plot_type']]  # a string defining a colon-separated list of axis names. An axis can be a channel name or "<idx>". This shares the syntax of the NeXus @axes attribute

    return ret



def getElementTypeIcon(t):
    if t == ChannelView.Channel:
        return getIcon(":/actions/system-shutdown.svg")
    elif t == ChannelView.Enabled:
        return getIcon(":/status/true.svg")
    elif t == ChannelView.Output:
        return getThemeIcon("utilities-terminal")
    elif t == ChannelView.PlotType:
        return getIcon(":/apps/utilities-system-monitor.svg")
    elif t == ChannelView.PlotAxes:
        return getIcon(":/apps/utilities-system-monitor.svg")
    elif t == ChannelView.Timer:
        return getIcon(":/status/flag-green-clock.svg")
    elif t == ChannelView.Monitor:
        return getIcon(":/status/flag-green.svg")
    elif t == ChannelView.Trigger:
        return getIcon(":/actions/system-shutdown.svg")
    elif t == ChannelView.NXPath:
        return getThemeIcon("document-save-as")

    return getIcon(":/tango.png")


def getElementTypeSize(t):
    if t == ChannelView.Channel:
        return Qt.QSize(200, 24)
    elif t == ChannelView.Enabled:
        return Qt.QSize(50, 24)
    elif t == ChannelView.Output:
        return Qt.QSize(50, 24)
    elif t == ChannelView.PlotType:
        return Qt.QSize(50, 24)
    return Qt.QSize(50, 24)


def getElementTypeToolTip(t):
    """Wrapper to prevent loading qtgui when this module is imported"""
    if t == ChannelView.Channel:
        return "Channel"
    elif t == ChannelView.Enabled:
        return "Channel active or not"
    elif t == ChannelView.Output:
        return "Channel output active or not"
    elif t == ChannelView.Shape:
        return "Shape of the data (using numpy convention). For example, a scalar will have shape=(), a spectrum of 10 elements will have shape=(10,) and an image of 20x30 will be shape=(20,30)"
    elif t == ChannelView.DataType:
        return "Type of data for storing (valid types are: char, float32, float64, [u]int{8|16|32|64})",
    elif t == ChannelView.PlotType:
        return "Plot type for this channel "
    elif t == ChannelView.PlotAxes:
        return "Independent variables to be used in the plot of this channel"
    elif t == ChannelView.Timer:
        return "The channel to be used as the timer"
    elif t == ChannelView.Monitor:
        return "The channel to be used as a monitor for stopping the acquisition"
    elif t == ChannelView.Trigger:
        return "The channel to be used for triggering the acquisition"
    elif t == ChannelView.Conditioning:
        return "An expression to evaluate on the data when displaying it"
    elif t == ChannelView.Normalization:
        return "Normalization mode for the data"
    elif t == ChannelView.NXPath:
        return "Location of the data of this channel within the NeXus tree"
    return "Unknown"


class BaseMntGrpChannelItem(TaurusBaseTreeItem):
    """ """
    def data(self, index):
        """Returns the data of this node for the given index

        :return: (object) the data for the given index
        """
        return self._itemData

    def role(self):
        """Returns the prefered role for the item.
        This implementation returns ChannelView.Unknown

        This method should be able to return any kind of python object as long
        as the model that is used is compatible.

        :return: (MacroView) the role in form of element type"""
        return ChannelView.Unknown


class MntGrpChannelItem(BaseMntGrpChannelItem):

    itemdata_keys_map = {ChannelView.Channel:'label',
                         ChannelView.Enabled:'enabled',
                         ChannelView.Output:'output',
                         ChannelView.Shape:'shape',
                         ChannelView.DataType:'data_type',
                         ChannelView.PlotType:'plot_type',
                         ChannelView.PlotAxes:'plot_axes',
#                         ChannelView.Timer:'timer',
#                         ChannelView.Monitor:'monitor',
#                         ChannelView.Trigger:'trigger',
                         ChannelView.Conditioning:'conditioning',
                         ChannelView.Normalization:'normalization',
                         ChannelView.NXPath:'nexus_path'
                         }

    def data(self, index):
        """Returns the data of this node for the given index

        :return: (object) the data for the given index
        """
        taurus_role = index.model().role(index.column())
        ch_name, ch_data = self.itemData()
        key = self.itemdata_keys_map[taurus_role]
        ret = ch_data[key]
        if taurus_role == ChannelView.PlotType:
            ret = PlotType[ret]
        elif taurus_role == ChannelView.Normalization:
            ret = Normalization[ret]
        elif taurus_role == ChannelView.PlotAxes:
            ret = ":".join(ret)
        elif taurus_role == ChannelView.Shape:
            ret = str(ret)
        return ret

    def setData(self, index, qvalue):
        taurus_role = index.model().role(index.column())
        str_value = Qt.from_qvariant(qvalue, str)
        if taurus_role in (ChannelView.Channel, ChannelView.Conditioning,
                           ChannelView.NXPath, ChannelView.DataType) :
            data = str_value
        elif taurus_role in (ChannelView.Enabled, ChannelView.Output) :
            data = Qt.from_qvariant(qvalue, bool)
        elif taurus_role == ChannelView.PlotType:
            data = PlotType[str_value]
        elif taurus_role == ChannelView.Normalization:
            data = Normalization[str_value]
        elif taurus_role == ChannelView.PlotAxes:
            data = [a for a in str_value.split(':')]
        elif taurus_role == ChannelView.Shape:
            s = str_value
            try:
                data = eval(s, {}, {})
                if not isinstance(data, (tuple, list)):
                    raise ValueError
            except:
                from taurus.core.util.log import Logger
                Logger(self.__class__.__name__).error('Invalid shape %s', s)
                data = ()
        else:
            raise NotImplementedError('Unknown role')
        ch_name, ch_data = self.itemData()
        key = self.itemdata_keys_map[taurus_role]
        ch_data[key] = data

    def role(self):
        return ChannelView.Channel

    def toolTip(self, index):
        return "Channel " + self._itemData[0]

    def icon(self, index):
        taurus_role = index.model().role(index.column())
        if taurus_role == ChannelView.Channel:
            return getIcon(":/actions/system-shutdown.svg")


class MntGrpUnitItem(TaurusBaseTreeItem):
    pass


class BaseMntGrpChannelModel(TaurusBaseModel):
    ColumnNames = ("Channel", "enabled", "output", "Shape", "Data Type", "Plot Type", "Plot Axes", "Timer",
                  "Monitor", "Trigger", "Conditioning", "Normalization", "NeXus Path")
    ColumnRoles = ((ChannelView.Channel, ChannelView.Channel), ChannelView.Enabled,
                  ChannelView.Output, ChannelView.Shape, ChannelView.DataType, ChannelView.PlotType,
                  ChannelView.PlotAxes, ChannelView.Timer, ChannelView.Monitor,
                  ChannelView.Trigger, ChannelView.Conditioning,
                  ChannelView.Normalization, ChannelView.NXPath)
    DftFont = Qt.QFont()

    _availableChannels = {}
    data_keys_map = {ChannelView.Timer:'timer',
                     ChannelView.Monitor:'monitor',
                     ChannelView.Trigger:'trigger_type',
                     }

    def __init__(self, parent=None, data=None):
        TaurusBaseModel.__init__(self, parent=parent, data=data)
        self._mgconfig = None
        self._dirty = False

    def setAvailableChannels(self, cdict):
        self._availableChannels = cdict

    def getAvailableChannels(self):
        return self._availableChannels

    def createNewRootItem(self):
        return BaseMntGrpChannelItem(self, self.ColumnNames)

    def roleIcon(self, taurus_role):
        return getElementTypeIcon(taurus_role)

    def roleSize(self, taurus_role):
        return getElementTypeSize(taurus_role)

    def roleToolTip(self, taurus_role):
        return getElementTypeToolTip(taurus_role)

    def getPyData(self, ctrlname=None, unitid=None, chname=None, key=None):
        '''
        If controller name, unitid and channel name are given, it returns the dictionary with the channel info.
        If only controller name and unit id are given, it returns the dictionary with the unit info.
        If only controller name is given, it returns the dictionary with the controller info.

        Note that it will raise a KeyError exception if any of the keys are not
        found or if chname is given without providing the unit id
        '''
        if ctrlname is None:
            raise ValueError('controller name must be passed')
        if unitid is None:
            return self._mgconfig['controllers'][ctrlname]
        elif chname is None:
            return  self._mgconfig['controllers'][ctrlname]['units'][unitid]
        else:
            return  self._mgconfig['controllers'][ctrlname]['units'][unitid]['channels'][chname]

    def setupModelData(self, mgconfig):
        if mgconfig is None:
            return
        root = self._rootItem  #@The root could eventually be changed for each unit or controller
        channelNodes = [MntGrpChannelItem(self, chcfg, root) for chcfg in getChannelConfigs(mgconfig)]
        for ch in channelNodes:
            root.appendChild(ch)
        self.updateMntGrpChannelIndex(root=root)
        #store the whole config object for accessing the info that is not at the channel level
        self._mgconfig = mgconfig

    def setDataSource(self, data_src):
        self._dirty = False
        TaurusBaseModel.setDataSource(self, data_src)

    def updateMntGrpChannelIndex(self, root=None):
        '''
        assigns the MeasurementGroup index (the internal order in the MG)
        according to the order in the QModel
        '''
        if root is None:
            root = self._rootItem
        for row in range(root.childCount()):
            chname, chdata = root.child(row).itemData()
            chdata['index'] = row

    def flags(self, index):
        flags = TaurusBaseModel.flags(self, index)
        taurus_role = self.role(index.column())
        if taurus_role == ChannelView.Channel:  #channel column is not editable
            return flags
        elif taurus_role == ChannelView.Trigger:
            ch_name, ch_data = index.internalPointer().itemData()
            if not ch_data['_controller_name'].startswith("__"):
                ch_info = self.getAvailableChannels()[ch_name]
                #only timer/monitor columns of counter timers are editable
                if ch_info['type'] in ('CTExpChannel', 'OneDExpChannel', 'TwoDExpChannel'):
                    flags |= Qt.Qt.ItemIsEditable
        elif taurus_role in (ChannelView.Timer, ChannelView.Monitor):
            ch_name, ch_data = index.internalPointer().itemData()
            if not ch_data['_controller_name'].startswith("__"):
                #ch_info = self.getAvailableChannels()[ch_name]
                #if 'CTExpChannel' == ch_info['type']: #only timer/monitor columns of counter timers are editable
                #    flags |= Qt.Qt.ItemIsEditable
                flags |= Qt.Qt.ItemIsEditable
        else:
            flags |= Qt.Qt.ItemIsEditable
        return flags

    def data(self, index, role=Qt.Qt.DisplayRole):
        """Reimplemented from :meth:`TaurusBaseModel.data`

        :return: (object) the data for the given index
        """
        #Try with the normal TaurusBaseModel item-oriented approach
        try:
            return TaurusBaseModel.data(self, index, role=role)
        except:
            pass
        #For those things which are inter-item, we handle them here
        taurus_role = self.role(index.column())
        if taurus_role == ChannelView.Trigger:
            ch_name, ch_data = index.internalPointer().itemData()
            unitdict = self.getPyData(ctrlname=ch_data['_controller_name'], unitid=ch_data['_unit_id'])
            key = self.data_keys_map[taurus_role]
            return Qt.QVariant(AcqTriggerType[unitdict.get(key, None)])
        elif taurus_role in (ChannelView.Timer, ChannelView.Monitor):
            ch_name, ch_data = index.internalPointer().itemData()
            ctrlname = ch_data['_controller_name']
            if ctrlname.startswith("__"):
                return Qt.QVariant()
            ch_info = self.getAvailableChannels()[ch_name]
            if ch_info['type'] in ('CTExpChannel', 'OneDExpChannel', 'TwoDExpChannel'):
                unitdict = self.getPyData(ctrlname=ctrlname, unitid=ch_data['_unit_id'])
                key = self.data_keys_map[taurus_role]
                master_full_name = unitdict.get(key, None)
            else:
                key = taurus_role == ChannelView.Timer and 'timer' or 'monitor'
                master_full_name = self._mgconfig.get(key, None)
            if master_full_name is None:
                return Qt.QVariant()
            else:
                master_info = self.getAvailableChannels()[master_full_name]
                return Qt.QVariant(master_info['name'])

        return Qt.QVariant()

    def setData(self, index, qvalue, role=Qt.Qt.EditRole):
        #For those things which are at the unit level, we handle them here
        taurus_role = self.role(index.column())
        if taurus_role in (ChannelView.Timer, ChannelView.Monitor, ChannelView.Trigger):
            ch_name, ch_data = index.internalPointer().itemData()
            ch_info = self.getAvailableChannels()[ch_name]
            unit_data = self.getPyData(ctrlname=ch_data['_controller_name'], unitid=ch_data['_unit_id'])
            key = self.data_keys_map[taurus_role]
            data = Qt.from_qvariant(qvalue, str)

            self._dirty = True
            self.beginResetModel()
            is_settable = ch_info['type'] in ('CTExpChannel', 'OneDExpChannel', 'TwoDExpChannel')
            if taurus_role == ChannelView.Trigger:
                data = AcqTriggerType[data]
                if is_settable:
                    unit_data[key] = data
            else:
                if is_settable:
                    if unit_data[key] == self._mgconfig[key]:
                        self._mgconfig[key] = data
                    unit_data[key] = data
                else:
                    self._mgconfig[key] = data
            self.endResetModel()
            return True
        #for the rest, we use the regular TaurusBaseModel item-oriented approach
        #ret = self._setData(index, qvalue, role) #@todo we do not use _setData because it is not Qt4.4-compatible
        item = index.internalPointer()
        item.setData(index, qvalue)
        self._dirty = True
        self.emit(Qt.SIGNAL("dataChanged(const QModelIndex &, const QModelIndex &)"),
                  index, index)
        return True

    def addChannel(self, chname=None, chinfo=None, ctrlname=None, unitname=None, external=False):  #@todo: Very inefficient implementation. We should use {begin|end}InsertRows

        if chname is None:
            chname = chinfo['full_name']
        chname = str(chname)
        if ctrlname is None:
            desc = self.getAvailableChannels()[chname]
            ctrlname = desc['controller']
        if unitname is None:
            desc = self.getAvailableChannels()[chname]
            unitname = desc.get('unit', '0')  #@fixme: at the moment of writing, the unit info cannot be obtained from desc

        #update the internal data
        self.beginResetModel()  #we are altering the internal data here, so we need to protect it
        ctrlsdict = self.dataSource()['controllers']
        if not ctrlsdict.has_key(ctrlname): ctrlsdict[ctrlname] = {'units':{}}
        unitsdict = ctrlsdict[ctrlname]['units']
        if not unitsdict.has_key(unitname):
            unitsdict[unitname] = unit = {'channels':{}}
            if not external and chinfo['type'] in ('CTExpChannel', 'OneDExpChannel', 'TwoDExpChannel'):
                unit['timer'] = chname
                unit['monitor'] = chname
                unit['trigger_type'] = AcqTriggerType.Software
        channelsdict = unitsdict[unitname]['channels']
        if channelsdict.has_key(chname):
            self.error('Channel "%s" is already in the measurement group. It will not be added again' % chname)
            return

        self._dirty = True
        if external:
            channelsdict[chname] = createChannelDict(chname)
        else:
            channelsdict[chname] = createChannelDict(chinfo)
        self.endResetModel()  #we are altering the internal data here, so we need to protect it
        self.refresh()  #note that another reset will be done here...

        #import pprint
        #pprint.pprint(self.dataSource())

    def removeChannels(self, chnames):  #@todo: Very inefficient implementation. We should use {begin|end}InsertRows
        #update the internal data
        self._dirty = True
        self.beginResetModel()  #we are altering the internal data here, so we need to protect it
        for chname in chnames:
            avail_channels = self.getAvailableChannels()
            if chname in avail_channels:
                desc = self.getAvailableChannels()[chname]
                ctrlname = desc['controller']
                unitname = desc.get('unit', '0')  #@fixme: at the moment of writing, the unit info cannot be obtained from desc
            else:
                #@todo: This assumes that if it is not in the list of avail_channels, it must be an external tango channel
                ctrlname = '__tango__'
                unitname = '0'
            try:
                self.dataSource()['controllers'][ctrlname]['units'][unitname]['channels'].pop(chname)
                try:
                    if not self.dataSource()['controllers'][ctrlname]['units'][unitname]['channels']:
                        self.dataSource()['controllers'][ctrlname]['units'].pop(unitname)
                        if not self.dataSource()['controllers'][ctrlname]['units']:
                            self.dataSource()['controllers'].pop(ctrlname)
                except:
                    self.error('error cleaning the data source dictionary')
            except:
                self.error('cannot find "%s" for removing' % chname)

        self.endResetModel()  #we are altering the internal data here, so we need to protect it
        self.refresh()  #note that another reset will be done here...

    def swapChannels(self, root, row1, row2):  #@todo: Very inefficient implementation. We should use {begin|end}MoveRows
        self._dirty = True
        n1, d1 = root.child(row1).itemData()
        n2, d2 = root.child(row2).itemData()
        d1['index'], d2['index'] = d2['index'], d1['index']
        self.debug("swapping %s with %s" % (n1, n2))
        self.refresh()

    def isDataChanged(self):
        return self._dirty

    def setDataChanged(self, datachanged):
        self._dirty = datachanged

class MntGrpChannelModel(BaseMntGrpChannelModel):
    '''A BaseMntGrpChannelModel that communicates with a MntGrp device for setting and reading the configuration
    '''

    def setDataSource(self, mg):
        if self._data_src is not None:
            Qt.QObject.disconnect(self._data_src, Qt.SIGNAL('configurationChanged'), self.configurationChanged)
        if mg is not None:
            Qt.QObject.connect(mg, Qt.SIGNAL('configurationChanged'), self.configurationChanged)
        BaseMntGrpChannelModel.setDataSource(self, mg)

    def configurationChanged(self):
        self.refresh()

    def setupModelData(self, mg):
        if mg is None:
            return
        BaseMntGrpChannelModel.setupModelData(self, self.getSourceData())

    def writeSourceData(self):
        mg = self.dataSource()
        if mg is not None and self._mgconfig is not None:
            mg.setConfiguration(self._mgconfig)

    def getSourceData(self):
        """Gets data from the dataSource"""
        mg = self.dataSource()
        if mg is not None:
            return mg.getConfiguration()

    def getLocalData(self):
        """Gets the local data (may be different from the one in the data source
        since it may have been modified by the user)"""
        return self._mgconfig


class AxesSelector(Qt.QWidget):
    def __init__(self, parent, n=0, choices=None):
        '''Shows n comboboxes populated with choices. If n is 0, it just shows a LineEdit instead'''
        Qt.QWidget.__init__(self, parent)
        self._n = n
        self._CBs = []
        self._LE = None
        l = Qt.QHBoxLayout(self)
        if self._n == 0:
            self._LE = Qt.QLineEdit()
            l.addWidget(self._LE)
        else:
            for i in range(n):
                cb = Qt.QComboBox()
                l.addWidget(cb)
                self._CBs.append(cb)
        if choices is not None:
            self.setChoices(choices)

    def setChoices(self, choices):
        for cb in self._CBs:
            cb.addItems(choices)

    def text(self):
        return ":".join(self.getCurrentChoices())

    def getCurrentChoices(self):
        if self._LE is None:
            return [str(cb.currentText()) for cb in self._CBs]
        else:
            return [str(self._LE.text())]

    def setCurrentChoices(self, choice):
        if self._LE is None:
            texts = str(choice).split(':')
            for t, cb in zip(texts[:len(self._CBs)], self._CBs):
                cb.setCurrentIndex(max(0, cb.findText(t)))
        else:
            self._LE.setText(str(choice))


class ChannelDelegate(Qt.QStyledItemDelegate):
    def createEditor(self, parent, option, index):
        model = index.model()
        taurus_role = model.role(index.column())
        if taurus_role in (ChannelView.Channel, ChannelView.PlotType, ChannelView.Normalization,
                           ChannelView.Timer, ChannelView.Monitor, ChannelView.Trigger):
            ret = Qt.QComboBox(parent)
        elif taurus_role == ChannelView.PlotAxes:
            item = index.internalPointer()
            ptype = item.itemData()[1]['plot_type']
            if ptype == PlotType.Spectrum:
                n = 1
            elif ptype == PlotType.Image:
                n = 2
            else:
                return None
            ret = AxesSelector(parent, n=n)
        else:
            ret = Qt.QStyledItemDelegate.createEditor(self, parent, option, index)
        ret.setAutoFillBackground(True)
        return ret

    def setEditorData(self, editor, index):
        model = index.model()
        dataSource = model.dataSource()
        taurus_role = model.role(index.column())
        if taurus_role == ChannelView.PlotType:
            editor.addItems(PlotType.keys())
            current = Qt.from_qvariant(model.data(index), str)
            editor.setCurrentIndex(editor.findText(current))
        elif taurus_role == ChannelView.Normalization:
            editor.addItems(Normalization.keys())
            current = Qt.from_qvariant(model.data(index), str)
            editor.setCurrentIndex(editor.findText(current))
        elif taurus_role in (ChannelView.Timer, ChannelView.Monitor):
            key = taurus_role == ChannelView.Timer and 'timer' or 'monitor'
            ch_name, ch_data = index.internalPointer().itemData()
            ctrl_filterlist = [ch_data['_controller_name']]
            ctrl_dict = getChannelConfigs(dataSource, ctrls=ctrl_filterlist)
            all_channels = model.getAvailableChannels()
            # if it is a timer capable type of element
            if all_channels[ch_name]['type'] in ('CTExpChannel', 'OneDExpChannel', 'TwoDExpChannel'):
                for full_name, channel_data in ctrl_dict:
                    editor.addItem(channel_data['name'], Qt.QVariant(full_name))
                current = Qt.from_qvariant(model.data(index), str)
                editor.setCurrentIndex(editor.findText(current))
            else:
                for ctrl_data in dataSource['controllers'].values():
                    for unit_data in ctrl_data['units'].values():
                        if key in unit_data:
                            channel = all_channels[unit_data[key]]
                            editor.addItem(channel['name'], Qt.QVariant(channel['full_name']))
                current = dataSource.get(key)  # current global timer/monitor
                editor.setCurrentIndex(editor.findData(Qt.QVariant(current)))
        elif taurus_role == ChannelView.Trigger:
            editor.addItems(AcqTriggerType.keys())
            current = Qt.from_qvariant(model.data(index), str)
            editor.setCurrentIndex(editor.findText(current))
        elif taurus_role == ChannelView.PlotAxes:
            selectables = ['<idx>', '<mov>'] + [n for n, d in getChannelConfigs(dataSource)]
            editor.setChoices(selectables)
            current = Qt.from_qvariant(model.data(index), str)
            editor.setCurrentChoices(current)
        else:
            Qt.QStyledItemDelegate.setEditorData(self, editor, index)

    def setModelData(self, editor, model, index):
        taurus_role = model.role(index.column())
        dataSource = model.dataSource()
        if taurus_role in (ChannelView.Channel, ChannelView.PlotType, ChannelView.Normalization):
            data = Qt.QVariant(editor.currentText())
            model.setData(index, data)
        elif taurus_role == ChannelView.Trigger:
            old_value = Qt.from_qvariant(model.data(index), str)
            new_value = str(editor.currentText())
            if new_value == old_value:
                return
            ch_name, ch_data = index.internalPointer().itemData()
            channels = getChannelConfigs(dataSource, ctrls=[ch_data['_controller_name']], units=[ch_data['_unit_id']])
            affected = [d['name'] for n, d in channels]
            if len(affected) > 1:
                op = Qt.QMessageBox.question(editor, "Caution: multiple channels affected",
                                            "This change will also affect the following channels:\n- %s \nContinue?" % "\n- ".join(affected),
                                            Qt.QMessageBox.Yes | Qt.QMessageBox.Cancel)
                if op != Qt.QMessageBox.Yes:
                    return
            data = Qt.QVariant(new_value)
            model.setData(index, data)
        elif taurus_role in (ChannelView.Timer, ChannelView.Monitor):
            key = taurus_role == ChannelView.Timer and 'timer' or 'monitor'
            old_value = Qt.from_qvariant(model.data(index), str)
            new_value = str(editor.currentText())
            if new_value == old_value:
                return
            ch_name, ch_data = index.internalPointer().itemData()
            all_channels = model.getAvailableChannels()
            # if it is a timer capable type of element
            ch_info = all_channels[ch_name]
            selected_master = editor.itemData(editor.currentIndex())
            if ch_info['type'] in ('CTExpChannel', 'OneDExpChannel', 'TwoDExpChannel'):
                affected = []
                unit_data = model.getPyData(ctrlname=ch_data['_controller_name'], unitid=ch_data['_unit_id'])
                channels = getChannelConfigs(dataSource, ctrls=[ch_data['_controller_name']], units=[ch_data['_unit_id']])
                for n, d in channels:
                    affected.append(d['name'])
                # if old timer/monitor was also the global, then non
                # timerable/monitorable channels must be changed
                if unit_data[key] == dataSource.get(key):
                    for n, d in getChannelConfigs(dataSource):
                        if d['_controller_name'].startswith("__"):
                            continue
                        ch_info = all_channels[n]
                        if ch_info['type'] in ('CTExpChannel', 'OneDExpChannel', 'TwoDExpChannel'):
                            continue
                        affected.append(d['name'])

                if len(affected) > 1:
                    op = Qt.QMessageBox.question(editor, "Caution: multiple channels affected",
                                                "This change will also affect the following channels:\n- %s \nContinue?" % "\n- ".join(affected),
                                                Qt.QMessageBox.Yes | Qt.QMessageBox.Cancel)
                    if op != Qt.QMessageBox.Yes:
                        return
            else:
                affected = []
                channels = getChannelConfigs(dataSource)
                for n, d in channels:
                    if d['_controller_name'].startswith("__"):
                        continue
                    ch_info = all_channels[n]
                    if ch_info['type'] in ('CTExpChannel', 'OneDExpChannel', 'TwoDExpChannel'):
                        continue
                    affected.append(d['name'])
                if len(affected) > 1:
                    op = Qt.QMessageBox.question(editor, "Caution: multiple channels affected",
                                                "This change will also affect the following channels:\n- %s \nContinue?" % "\n- ".join(affected),
                                                Qt.QMessageBox.Yes | Qt.QMessageBox.Cancel)
                    if op != Qt.QMessageBox.Yes:
                        return
            model.setData(index, selected_master)
        elif taurus_role == ChannelView.PlotAxes:
            data = Qt.QVariant(editor.text())
            model.setData(index, data)
        else:
            Qt.QStyledItemDelegate.setModelData(self, editor, model, index)


class MntGrpChannelEditor(TaurusBaseTableWidget):
    """
    """

    KnownPerspectives = {
        "Channel" : {
            "label"   : "Channels",
            "icon"    : ":/actions/system-shutdown.svg",
            "tooltip" : "View by channel",
            "model"   : [BaseMntGrpChannelModel, ],
        },
    }

    DftPerspective = "Channel"
    _simpleViewColumns = (ChannelView.Channel, ChannelView.Output, ChannelView.Shape, ChannelView.PlotType, ChannelView.PlotAxes)
    _simpleView = False

    def __init__(self, parent=None, designMode=False, with_filter_widget=True, perspective=None):
        TaurusBaseTableWidget.__init__(self, parent=parent, designMode=designMode,
                                       with_filter_widget=with_filter_widget,
                                       perspective=perspective, proxy=None)
        self.setContextMenuPolicy(Qt.Qt.ActionsContextMenu)
        self._simpleViewAction = Qt.QAction("Simple View", self)
        self._simpleViewAction.setCheckable(True)
        self.connect(self._simpleViewAction, Qt.SIGNAL("toggled(bool)"), self.setSimpleView)
        self.addAction(self._simpleViewAction)
        self.registerConfigProperty(self.isSimpleView, self.setSimpleView, "simpleView")

    def isSimpleView(self):
        return self._simpleView

    def setSimpleView(self, simpleview):
        if simpleview == self.isSimpleView():
            return
        columnRoles = list(self.getQModel().ColumnRoles)
        columnRoles[0] = columnRoles[0][-1]  #account for the fact that the first element is a tuple instead of a role
        columnIndexes = [columnRoles.index(r) for r in self._simpleViewColumns]
        for i in range(self.getQModel().columnCount()):
            hide = simpleview and (i not in columnIndexes)
            self.tableView().setColumnHidden(i, hide)
        self._simpleView = simpleview
        self._simpleViewAction.setChecked(simpleview)

    def resetSimpleView(self):
        self.setSimpleView(False)

    def createViewWidget(self):
        tableView = TaurusBaseTableWidget.createViewWidget(self)
        self._delegate = ChannelDelegate(self)
        #self._delegate.setItemEditorFactory(Qt.QItemEditorFactory()) #This causes a segfault when calling ChannelDelegate.createEditor
        tableView.setItemDelegate(self._delegate)
        tableView.setSortingEnabled(False)
        self.connect(self._editorBar, Qt.SIGNAL("addTriggered"), self.addChannel)
        self.connect(self._editorBar, Qt.SIGNAL("removeTriggered"), self.removeChannels)
        self.connect(self._editorBar, Qt.SIGNAL("moveUpTriggered"), self.moveUpChannel)
        self.connect(self._editorBar, Qt.SIGNAL("moveDownTriggered"), self.moveDownChannel)
        return tableView

    def createToolArea(self):
        ta = TaurusBaseTableWidget.createToolArea(self)
        e_bar = self._editorBar = EditorToolBar(self, self)
        ta.append(e_bar)
        return ta

    def getModelClass(self):
        return TaurusDevice

    def addChannel(self, channel=None):
        qmodel = self.getQModel()
        dataSource = qmodel.dataSource()
        if channel is None:
            shown = [n for n, d in getChannelConfigs(dataSource)]
            avail_channels = qmodel.getAvailableChannels()
            clist = [ ch_info['name'] for ch_name, ch_info in avail_channels.items()
                      if ch_name not in shown ]
            clist = sorted(clist) + ['(Other...)']
            chname, ok = Qt.QInputDialog.getItem(self, "New Channel", "Choose channel:", clist, 0, False)
            if not ok:
                return
        chname = str(chname)
        if chname == '(Other...)':
            models, ok = TaurusModelChooser.modelChooserDlg(parent=self, singleModel=False, windowTitle='Choose source of data',
                                                            selectables=[TaurusElementType.Attribute])
            if not ok:
                return
            for m in models:
                qmodel.addChannel(chname=m, ctrlname='__tango__', unitname='0', external=True)
        else:
            for ch_info in avail_channels.values():
                if ch_info['name'] == chname:
                    qmodel.addChannel(chinfo=ch_info)

    def removeChannels(self, channels=None):
        if channels is None:
            channels = self.selectedItems()
        chnames = [ch.itemData()[0] for ch in channels]
        self.getQModel().removeChannels(chnames)

    def moveUpChannel(self, channel=None):
        if channel is None:
            channels = self.selectedItems()
            if len(channels) != 1:
                return
            channel = channels[0]
        parent = channel.parent()
        row = channel.row()
        if row < 1:
            return

        model = self.getQModel()
        model.swapChannels(parent, row, row - 1)
        idx = model.index(row - 1, 0)
        self.viewWidget().setCurrentIndex(idx)

    def moveDownChannel(self, channel=None):
        if channel is None:
            channels = self.selectedItems()
            if len(channels) != 1:
                return
            channel = channels[0]
        parent = channel.parent()
        row = channel.row()
        if row >= parent.childCount() - 1:
            return
        model = self.getQModel()
        self.getQModel().swapChannels(parent, row, row + 1)
        idx = model.index(row + 1, 0)
        self.viewWidget().setCurrentIndex(idx)

    def getLocalConfig(self):
        return self.getQModel().getLocalData()

    @classmethod
    def getQtDesignerPluginInfo(cls):
        ret = TaurusBaseTableWidget.getQtDesignerPluginInfo()
        ret['module'] = 'taurus.qt.qtgui.extra_sardana'
        ret['group'] = 'Taurus Sardana'
        ret['icon'] = ":/designer/table.png"
        return ret

    simpleView = Qt.pyqtProperty("bool", isSimpleView, setSimpleView, resetSimpleView)





##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This is the standard macro module"""

__docformat__ = 'restructuredtext'

__all__ = ["get", "put"]

from sardana.macroserver.macro import *

import array

class put(Macro):
    """Sends a string to the communication channel"""
    param_def = [
       ['communication channel', Type.ComChannel, None, 'the communication channel'],
       ['data', Type.String, None, 'data to be sent']
    ]
    
    def run(self, comch, data):
        name = comch.getName()
        nb = comch.write(data)
        o = str(nb) + " bytes sent to " + name
        self.output(o)


class get(Macro):
    """Reads and outputs the data from the communication channel"""
    
    param_def = [
       ['communication channel', Type.ComChannel, None, 'the communication channel'],
       ['maximum length', Type.String, -1, 'maximum number of bytes to read']
    ]
    
    def run(self, comch, maxlen):
        name = comch.getName()
        data = comch.command_inout("read",maxlen)
        
        datastr = array.array('B',data).tostring()
        self.output("'" + datastr + "'")
        self.output(data)


#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""
favouriteseditor.py: 
"""
import copy

from taurus.external.qt import Qt
from taurus.qt.qtgui.container import TaurusWidget
from taurus.qt.qtcore.configuration import BaseConfigurableClass
from model import MacrosListModel

class FavouritesMacrosEditor(TaurusWidget):
    __pyqtSignals__ = ("modelChanged(const QString &)",)

    def __init__(self, parent=None, designMode=False):
        TaurusWidget.__init__(self, parent, designMode)
        self.setObjectName(self.__class__.__name__)
        self.registerConfigProperty("toXmlString", "fromXmlString", "favourites")
        self.initComponents()

    def initComponents(self):
        self.setLayout(Qt.QHBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)

        self.list = FavouritesMacrosList(self)
        self._model = MacrosListModel()
        self.list.setModel(self._model)

#        self.registerConfigDelegate(self.list)
        self.layout().addWidget(self.list)

        actionBar = self.createActionBar()
        self.layout().addLayout(actionBar)

    def createActionBar(self):
        layout = Qt.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        deleteButton = Qt.QToolButton()
        deleteButton.setDefaultAction(self.list.removeAction)
        layout.addWidget(deleteButton)
        deleteAllButton = Qt.QToolButton()
        deleteAllButton.setDefaultAction(self.list.removeAllAction)
        layout.addWidget(deleteAllButton)
        moveUpButton = Qt.QToolButton()
        moveUpButton.setDefaultAction(self.list.moveUpAction)
        layout.addWidget(moveUpButton)
        moveDownButton = Qt.QToolButton()
        moveDownButton.setDefaultAction(self.list.moveDownAction)
        layout.addWidget(moveDownButton)
        spacerItem = Qt.QSpacerItem(0, 0, Qt.QSizePolicy.Fixed, Qt.QSizePolicy.Expanding)
        layout.addItem(spacerItem)
        return layout

    def addMacro(self, macroNode):
        self.list.insertMacro(macroNode)

    def toXmlString(self):
        return self.list.toXmlString()

    def fromXmlString(self, xmlString):
        self.list.fromXmlString(xmlString)
        favouritesList = self.list.model().list
        macroServerObj = self.getModelObj()
        if macroServerObj is None:
            self.debug("MS IS NONE")
            return

        for macroNode in favouritesList:
            macroServerObj.fillMacroNodeAdditionalInfos(macroNode)

    @classmethod
    def getQtDesignerPluginInfo(cls):
        return None


class FavouritesMacrosList(Qt.QListView, BaseConfigurableClass):

    def __init__(self, parent):
        Qt.QListView.__init__(self, parent)

        self.setSelectionMode(Qt.QListView.ExtendedSelection)

        self.removeAction = Qt.QAction(Qt.QIcon(":/actions/list-remove.svg"), "Remove from favourites", self)
        self.connect(self.removeAction, Qt.SIGNAL("triggered()"), self.removeMacros)
        self.removeAction.setToolTip("Clicking this button will remov selected macros from favourites.")

        self.removeAllAction = Qt.QAction(Qt.QIcon(":/places/user-trash.svg"), "Remove all from favourites", self)
        self.connect(self.removeAllAction, Qt.SIGNAL("triggered()"), self.removeAllMacros)
        self.removeAllAction.setToolTip("Clicking this button will remove all macros from favourites.")

        self.moveUpAction = Qt.QAction(Qt.QIcon(":/actions/go-up.svg"), "Move up", self)
        self.connect(self.moveUpAction, Qt.SIGNAL("triggered()"), self.upMacro)
        self.moveUpAction.setToolTip("Clicking this button will move the macro up in the favourites hierarchy.")

        self.moveDownAction = Qt.QAction(Qt.QIcon(":/actions/go-down.svg"), "Move up", self)
        self.connect(self.moveDownAction, Qt.SIGNAL("triggered()"), self.downMacro)
        self.moveDownAction.setToolTip("Clicking this button will move the macro down in the favourites hierarchy.")

        self.disableActions()

    def currentChanged(self, current, previous):
        macro = copy.deepcopy(self.currentIndex().internalPointer())
        self.emit(Qt.SIGNAL("favouriteSelected"), macro)
        Qt.QListView.currentChanged(self, current, previous)

    def selectionChanged(self, old, new):
        macro = None
        if self.currentIndex().isValid():
            self.removeAllAction.setEnabled(True)
            self.isIndexSelected()
        else:
            self.disableActions()
        Qt.QListView.selectionChanged(self, old, new)
        if len(self.selectedIndexes()) > 1:
            self.moveUpAction.setEnabled(False)
            self.moveDownAction.setEnabled(False)

    def isIndexSelected(self):
        if len(self.selectedIndexes()) > 0:
            self.removeAction.setEnabled(True)
            self.moveUpAction.setEnabled(self.model().isUpRowAllowed(self.currentIndex()))
            self.moveDownAction.setEnabled(self.model().isDownRowAllowed(self.currentIndex()))
        else:
            self.removeAction.setEnabled(False)
            self.moveUpAction.setEnabled(False)
            self.moveDownAction.setEnabled(False)

    def mousePressEvent(self, e):
        clickedIndex = self.indexAt(e.pos())
        if clickedIndex.isValid():
            macro = copy.deepcopy(self.currentIndex().internalPointer())
            self.emit(Qt.SIGNAL("favouriteSelected"), macro)
        Qt.QListView.mousePressEvent(self, e)

    def disableActions(self):
        self.removeAction.setEnabled(False)
        self.removeAllAction.setEnabled(False)
        self.moveUpAction.setEnabled(False)
        self.moveDownAction.setEnabled(False)

    def insertMacro(self, macroNode):
        idx = self.model().insertRow(macroNode)
        self.setCurrentIndex(idx)

    def removeMacros(self):
        slist = sorted(self.selectedIndexes(), key=lambda index: index.row(), reverse=True)
        for index in slist:
            row = index.row()
            idx = self.model().removeRow(row)
        self.setCurrentIndex(idx)

    def removeAllMacros(self):
        self.selectAll()
        slist = sorted(self.selectedIndexes(), key=lambda index: index.row(), reverse=True)
        for index in slist:
            self.model().removeRow(index.row())

    def upMacro(self):
        row = self.currentIndex().row()
        idx = self.model().upRow(row)
        self.setCurrentIndex(idx)

    def downMacro(self):
        row = self.currentIndex().row()
        idx = self.model().downRow(row)
        self.setCurrentIndex(idx)

    def toXmlString(self):
        return self.model().toXmlString()

    def fromXmlString(self, xmlString):
        self.model().fromXmlString(xmlString)


def test():
    import sys, taurus, time
    from  taurus.qt.qtgui.application import TaurusApplication

    app = TaurusApplication(sys.argv)

    favouritesEditor = FavouritesMacrosEditor()

    args = app.get_command_line_args()
    favouritesEditor.setModel(args[0])
    time.sleep(1)
    macroNode = favouritesEditor.getModelObj().getMacroNodeObj(str(args[1]))
    favouritesEditor.addMacro(macroNode)
    favouritesEditor.show()

    sys.exit(app.exec_())

if __name__ == "__main__":
    test()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

##############################################################################
##
## This file is part of Sardana
## 
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""helper methods for sardana sphinx documentation"""

__expr = ('or',)

def process_type(t, obj_type='class'):
    t = t.strip()
    if not t: return ''
    if t in __expr: return t
    if t.count(' or '):
       i = t.index(' or ')
       return ' '.join(map(process_type, (t[:i],'or', t[i+4:])))
    if not t.count('<') or not t.count('>'): return ':%s:`%s`' % (obj_type, t)
    
    #process a container template
    start, stop = t.index('<'), t.index('>')
    main_type = t[:start]
    main_type = process_type(main_type)
    types = t[start+1:stop].split(',')
    types = ', '.join(map(process_type, types))
    return "%s <%s>" % (main_type, types)

def process_param(line):
    new_lines = []
    try:
        prefix, param, desc = line.split(':', 2)
        p, param_name = param.split()
        desc = desc.strip()
        if desc[0] == '(' :
            pos = desc.find(')')
            if pos != -1:
                elem_type = desc[1:pos]
                klass = process_type(elem_type)
                desc = desc[pos+1:]
                new_lines.append('%s:type %s: %s' % (prefix, param_name, klass))
        new_lines.append('%s:param %s: %s' % (prefix, param_name, desc))
    except Exception, e:
        print "Sardana sphinx extension: Not able to process param: '%s'" % line
        print "      Reason:",str(e)
        new_lines.append(line)
    return new_lines

def process_return(line):
    new_lines = []
    try:
        prefix, param, desc = line.split(':', 2)
        desc = desc.strip()
        if desc[0] == '(' :
            pos = desc.find(')')
            if pos != -1:
                elem_type = desc[1:pos]
                klass = process_type(elem_type)
                desc = desc[pos+1:]
                new_lines.append('%s:rtype: %s' % (prefix, klass))
        new_lines.append('%s:return: %s' % (prefix, desc))
    except Exception, e:
        print "Sardana sphinx extension: Not able to process 'return': '%s'" % line
        print "      Reason:",str(e)
        new_lines.append(line)
    return new_lines

def process_raise(line):
    new_lines = []
    try:
        prefix, param, desc = line.split(':', 2)
        desc = desc.strip()
        klass = ''
        if desc[0] == '(' :
            pos = desc.find(')')
            if pos != -1:
                elem_type = desc[1:pos]
                klass = "(" + process_type(elem_type, obj_type='exc') + ")"
                desc = desc[pos+1:]
        new_lines.append('%s:raise: %s %s' % (prefix, klass, desc))
    except Exception, e:
        print "Sardana sphinx extension: Not able to process 'raise': '%s'" % line
        print "      Reason:", str(e)
        new_lines.append(line)
    return new_lines

def _is_return(line):
    ret = line.startswith(':return')
    ret |= line.startswith(':returns')
    return ret

def _is_param(line):
    ret = line.startswith(':param')
    ret |= line.startswith(':parameter')
    ret |= line.startswith(':arg')
    ret |= line.startswith(':argument')
    ret |= line.startswith(':key')
    ret |= line.startswith(':keyword')
    return ret

def _is_raise(line):
    ret = line.startswith(':raise')
    ret |= line.startswith(':except')
    return ret

def process_docstring(app, what, name, obj, options, lines):
    ret = []
    for nb, line in enumerate(lines):
        line_strip = line.strip()
        if _is_param(line_strip):
            ret.extend(process_param(line))
        elif _is_return(line_strip):
            ret.extend(process_return(line))
        elif _is_raise(line_strip):
            ret.extend(process_raise(line))
        else:
            ret.append(line)
    
    del lines[:]
    lines.extend(ret)

import inspect
from sphinx.util.inspect import getargspec

def _format_method_args(obj):
    if inspect.isbuiltin(obj) or \
           inspect.ismethoddescriptor(obj):
        # can never get arguments of a C function or method
        return None
    argspec = getargspec(obj)
    if argspec[0] and argspec[0][0] in ('cls', 'self'):
        del argspec[0][0]
    return inspect.formatargspec(*argspec)

def _format_function_args(obj):
    if inspect.isbuiltin(obj) or \
           inspect.ismethoddescriptor(obj):
        # cannot introspect arguments of a C function or method
        return None
    try:
        argspec = getargspec(obj)
    except TypeError:
        # if a class should be documented as function (yay duck
        # typing) we try to use the constructor signature as function
        # signature without the first argument.
        try:
            argspec = getargspec(obj.__new__)
        except TypeError:
            argspec = getargspec(obj.__init__)
            if argspec[0]:
                del argspec[0][0]
    args = inspect.formatargspec(*argspec)
    # escape backslashes for reST
    args = args.replace('\\', '\\\\')
    return args

def process_signature(app, what, name, obj, options, signature, return_annotation):
    if hasattr(obj, "__wrapped__"):
        if what == "method":
            from taurus.core.util.wrap import wrapped
            obj = wrapped(obj)
            signature = _format_method_args(obj)
            return signature, return_annotation
                
    
def setup(app):
    #app.connect('autodoc-process-docstring', process_docstring)
    app.connect('autodoc-process-signature', process_signature)



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module is part of the Python Sardana libray. It defines the base classes
for sardana exceptions"""

from __future__ import absolute_import

__all__ = [ "AbortException", "SardanaException", "SardanaExceptionList",
           "UnknownCode", "UnknownLibrary", "LibraryError",
           "format_exception_only", "format_exception_only_str"]

__docformat__ = 'restructuredtext'

import sys
import traceback


def format_exception_only(etype, value):
    msg = traceback.format_exception_only(etype, value)
    if msg[-1].endswith("\n"):
        msg[-1] = msg[-1][:-1]
    return msg

def format_exception_only_str(etype, value):
    return "".join(format_exception_only(etype, value))

class AbortException(Exception):
    pass


class SardanaException(Exception):

    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args)
        if args:
            a1 = args[0]
            if isinstance(a1, dict):
                self.msg = a1.get("message", a1.get("msg", None))
                self.traceback = a1.get("traceback", a1.get("tb", None))
                self.type = a1.get("type", self.__class__.__name__)
            else:
                self.msg = str(a1)
                self.traceback = None
                self.type = self.__class__.__name__
        else:
            exc_info = kwargs.get("exc_info")
            if exc_info is None:
                self.msg = "Unknown sardana exception"
            else:
                msg = format_exception_only_str(*exc_info[:2])
                self.msg = msg
            self.traceback = None
            self.type = self.__class__.__name__
        if 'exc_info' in kwargs:
            self.exc_info = kwargs['exc_info']
        else:
            self.exc_info = kwargs.get('exc_info', sys.exc_info())

    def __str__(self):
        return self.msg

    def __repr__(self):
        return "%s: %s" % (self.type, self.msg)


class SardanaExceptionList(SardanaException):
    def __init__(self, *args):
        SardanaException.__init__(self, *args)
        self.exceptions = args[0]


class UnknownCode(SardanaException):
    pass


class UnknownLibrary(SardanaException):
    pass


class LibraryError(SardanaException):
    pass




#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""
macrobutton.py:
"""

__all__ = ['MacroButton']

import functools

import PyTango

import taurus
from  taurus.core import TaurusEventType
from taurus.external.qt import Qt
from taurus.qt.qtgui.container import TaurusWidget
from taurus.qt.qtgui.dialog import ProtectTaurusMessageBox
from taurus.core.util.colors import DEVICE_STATE_PALETTE
from taurus.qt.qtgui.util.ui import UILoadable


class DoorStateListener(Qt.QObject):

    __pyqtSignals__ = ["doorStateChanged"]

    def eventReceived(self, evt_src, evt_type, evt_value):
        if evt_type not in (TaurusEventType.Change, TaurusEventType.Periodic):
            return
        door_state = evt_value.value
        self.emit(Qt.SIGNAL('doorStateChanged'), door_state)

        
@UILoadable(with_ui='ui')
class MacroButton(TaurusWidget):
    ''' This class is intended to be used as a button to execute macros.
    The model must be a valid door.
    
    ..note:: Not implemented but will be needed: set an icon

    ..note::
        
        It may be useful to have a slot update_macro_param that any qwidget signal may be
        connected to. The only condition might be that the sender widget should have an
        attribute 'widget.macro_param_index' in order to determine where the value has
        to be added.
 
    ..note::
        
        It may be useful to have all the streams from qdoor available somehow
        (right-click?)
    '''

    __pyqtSignals__ = ['statusUpdated', 'resultUpdated']

    def __init__(self, parent=None, designMode=False):
        TaurusWidget.__init__(self, parent, designMode)
        self.loadUi()
        self.door = None
        self.door_state_listener = None
        self.macro_name = ''
        self.macro_args = []
        self.running_macro = None

        self.ui.progress.setValue(0)

        self.ui.button.setCheckable(True)
        self.connect(self.ui.button, Qt.SIGNAL('clicked()'), self.button_clicked)

    def toggleProgress(self, visible):
        self.ui.progress.setVisible(visible)

    def setModel(self, model):
        TaurusWidget.setModel(self, model)
        if self.door is not None:
            self.disconnect(self.door, Qt.SIGNAL('macroStatusUpdated'), self.statusUpdated)
            self.disconnect(self.door, Qt.SIGNAL('resultUpdated'), self.resultUpdated)
            self.door.getAttribute('State').removeListener(self.door_state_listener)

        try: self.door = taurus.Device(model)
        except: return

        self.connect(self.door, Qt.SIGNAL('macroStatusUpdated'), self.statusUpdated)
        self.connect(self.door, Qt.SIGNAL('resultUpdated'), self.resultUpdated)

        # Manage Door Tango States
        self.door_state_listener = DoorStateListener()
        self.connect(self.door_state_listener, Qt.SIGNAL('doorStateChanged'), self.doorStateChanged)
        self.door.getAttribute('State').addListener(self.door_state_listener)

    def doorStateChanged(self, state):
        color = '#' + DEVICE_STATE_PALETTE.hex(state)
        stylesheet = 'QFrame{border: 4px solid %s;}' % color
        self.ui.frame.setStyleSheet(stylesheet)

        # In case state is not ON, and macro not triggered by the button, disable it
        door_available = True
        if state not in [PyTango.DevState.ON, PyTango.DevState.ALARM] and not self.ui.button.isChecked():
            door_available = False
            self.ui.progress.setValue(0)

        self.ui.button.setEnabled(door_available)
        self.ui.progress.setEnabled(door_available)


    def statusUpdated(self, *args):
        # SHOULD SEE THE DOCUMENTATION ABOUT THE ARGS AND ALSO THE STATUS STATE MACHINE
        # ARGS FORMAT IS (GUESSING WITH PRINT STATEMENTS)
        # e.g. ((<sardana.taurus.core.tango.sardana.macro.Macro object at 0x7f29300bc210>, [{u'step': 100.0, u'state': u'stop', u'range': [0.0, 100.0], u'id': u'b226f5e8-c807-11e0-8abe-001d0969db5b'}]),)
        # ( (MacroObj, [status_dict, .?.]), .?.)

        # QUESTIONS: THIS MACRO OBJECT HAS ALOS STEP, RANGE, ...
        # AND ALSO THE STATUS DICT... WHICH SHOULD I USE?

        first_tuple = args[0]
        self.running_macro = first_tuple[0]

        status_dict = first_tuple[1][0]
        # KEYS RECEIVED FROM A 'SCAN' MACRO AND A 'TWICE' MACRO: IS IT GENERAL ?!?!?!
        state = status_dict['state']
        step = status_dict['step']
        step_range = status_dict['range']
        macro_id = status_dict['id']

        # Update progress bar
        self.ui.progress.setMinimum(step_range[0])
        self.ui.progress.setMaximum(step_range[1])
        self.ui.progress.setValue(step)

        if state in ['stop', 'abort', 'finish', 'alarm']:
            self.ui.button.setChecked(False)

        self.emit(Qt.SIGNAL('statusUpdated'), status_dict)

    def resultUpdated(self, *args):
        # ARGS APPEAR TO BE EMPTY... SHOULD THEY CONTAIN THE RESULT ?!?!?!
        # I have to rely on the 'macro object' received in the last status update
        if self.running_macro is None:
            return
        result = self.running_macro.getResult()
        self.emit(Qt.SIGNAL('resultUpdated'), result)

    def setText(self, text):
        self.setButtonText(text)

    def setButtonText(self, text):
        # SHOULD ALSO BE POSSIBLE TO SET AN ICON
        self.ui.button.setText(text)

    def setMacroName(self, macro_name):
        self.macro_name = str(macro_name)

    def updateMacroArgument(self, index, value):
        while len(self.macro_args) < index + 1:
            self.macro_args.append('')

        self.macro_args[index] = str(value)

    def updateMacroArgumentFromSignal(self, index, obj, signal):
        self.connect(obj, signal, functools.partial(self.updateMacroArgument, index))

    def button_clicked(self):
        if self.ui.button.isChecked():
            self.runMacro()
        else:
            self.abort()

    @ProtectTaurusMessageBox(msg='An error occurred trying to execute the macro.')
    def runMacro(self):
        if self.door is None:
            return

        # Thanks to gjover for the hint... :-D
        #macro_cmd = self.macro_name + ' ' + ' '.join(self.macro_args)
        macro_cmd_xml = '<macro name="%s">\n' % self.macro_name
        for arg in self.macro_args:
            macro_cmd_xml += '<param value="%s"/>\n' % arg
        macro_cmd_xml += '</macro>'
        try:
            #self.door.runMacro(macro_cmd)
            self.door.runMacro(macro_cmd_xml)
        except Exception, e:
            self.ui.button.setChecked(False)
            raise e

    def abort(self):
        if self.door is None:
            return
        self.door.PauseMacro()
        # Since this could be done by error (impatient users clicking more than once)
        # we provide a warning message that does not make the process too slow
        # It may also be useful and 'ABORT' at TaurusApplication level (macros+motions+acquisitions)
        title = 'Aborting macro'
        message = 'The following macro is still running:\n\n'
        message += '%s %s\n\n' % (self.macro_name, ' '.join(self.macro_args))
        message += 'Are you sure you want to abort?\n'
        buttons = Qt.QMessageBox.Ok | Qt.QMessageBox.Cancel
        ans = Qt.QMessageBox.warning(self, title, message, buttons, Qt.QMessageBox.Ok)
        if ans == Qt.QMessageBox.Ok:
            self.door.abort(synch=True)
        else:
            self.ui.button.setChecked(True)
            self.door.ResumeMacro()

    @classmethod
    def getQtDesignerPluginInfo(cls):
        return {'container': False,
                'group': 'Taurus Sardana',
                'module': 'taurus.qt.qtgui.extra_macroexecutor',
                'icon': ':/designer/pushbutton.png'}



class MacroButtonAbortDoor(TaurusWidget):
    def __init__(self, parent=None, designMode=False):
        TaurusWidget.__init__(self, parent, designMode)
        self.setLayout(Qt.QGridLayout())
        self.layout().setMargin(0)
        self.layout().setSpacing(0)

        self.btn_abort = Qt.QPushButton('Abort')
        sizePolicy = Qt.QSizePolicy(Qt.QSizePolicy.Expanding, Qt.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        self.btn_abort.setSizePolicy(sizePolicy)

        self.door = None
        self.btn_abort.setToolTip('Abort Macro')

        self.layout().addWidget(self.btn_abort, 0, 0)
        self.connect(self.btn_abort, Qt.SIGNAL('clicked()'), self.abort)

    def setModel(self, model):
        TaurusWidget.setModel(self, model)
        try: self.door = taurus.Device(model)
        except: self.door = None

    @ProtectTaurusMessageBox(msg='An error occurred trying to abort the macro.')
    def abort(self):
        if self.door is not None:
            self.door.stopMacro()



if __name__ == '__main__':
    import sys
    app = Qt.QApplication(sys.argv)

    w = Qt.QWidget()
    w.setLayout(Qt.QGridLayout())

    col = 0
    clear_button = Qt.QPushButton('clear')
    w.layout().addWidget(clear_button, 0, col, 2, 1)

    col += 1
    w.layout().addWidget(Qt.QLabel('macro name'), 0, col)
    macro_name = Qt.QLineEdit()
    w.layout().addWidget(macro_name, 1, col)

    col += 1
    w.layout().addWidget(Qt.QLabel('arg0'), 0, col)
    arg0 = Qt.QLineEdit()
    w.layout().addWidget(arg0, 1, col)

    col += 1
    w.layout().addWidget(Qt.QLabel('arg1'), 0, col)
    arg1 = Qt.QLineEdit()
    w.layout().addWidget(arg1, 1, col)

    col += 1
    w.layout().addWidget(Qt.QLabel('arg2'), 0, col)
    arg2 = Qt.QLineEdit()
    w.layout().addWidget(arg2, 1, col)

    col += 1
    w.layout().addWidget(Qt.QLabel('arg3'), 0, col)
    arg3 = Qt.QLineEdit()
    w.layout().addWidget(arg3, 1, col)

    col += 1
    w.layout().addWidget(Qt.QLabel('arg4'), 0, col)
    arg4 = Qt.QLineEdit()
    w.layout().addWidget(arg4, 1, col)

    from sardana.taurus.qt.qtcore.tango.sardana.macroserver import registerExtensions
    registerExtensions()
    mb = MacroButton()
    mb.setModel('door/gc/1')

    w.layout().addWidget(mb, 2, 0, 2, 7)

    w.layout().addWidget(Qt.QLabel('Result:'), 4, 0)

    result_label = Qt.QLabel()
    w.layout().addWidget(result_label, 4, 1, 1, 5)

    show_progress = Qt.QCheckBox('Progress')
    show_progress.setChecked(True)
    w.layout().addWidget(show_progress, 5, 0)

    mb_abort = MacroButtonAbortDoor()
    mb_abort.setModel('door/gc/1')
    w.layout().addWidget(mb_abort, 5, 1)

    # Change macro name
    Qt.QObject.connect(macro_name, Qt.SIGNAL('textChanged(QString)'), mb.setMacroName)
    Qt.QObject.connect(macro_name, Qt.SIGNAL('textChanged(QString)'), mb.setButtonText)

    # Change Nth macro argument
    mb.updateMacroArgumentFromSignal(0, arg0, Qt.SIGNAL('textChanged(QString)'))
    mb.updateMacroArgumentFromSignal(1, arg1, Qt.SIGNAL('textChanged(QString)'))
    mb.updateMacroArgumentFromSignal(2, arg2, Qt.SIGNAL('textChanged(QString)'))
    mb.updateMacroArgumentFromSignal(3, arg3, Qt.SIGNAL('textChanged(QString)'))
    mb.updateMacroArgumentFromSignal(4, arg4, Qt.SIGNAL('textChanged(QString)'))


    def update_result(result):
        result_label.setText(str(result))

    def toggle_progress(showProgress):
        visible = show_progress.isChecked()
        mb.toggleProgress(visible)

    def clear_params():
        for line_edit in [macro_name, arg0, arg1, arg2, arg3, arg4]:
            line_edit.setText('')

    # Toggle progressbar
    Qt.QObject.connect(show_progress, Qt.SIGNAL('stateChanged(int)'), toggle_progress)
    # Update possible macro result
    Qt.QObject.connect(mb, Qt.SIGNAL('resultUpdated'), update_result)
    # Clear parameters
    Qt.QObject.connect(clear_button, Qt.SIGNAL('clicked()'), clear_params)

    # Since everything is now connected, the parameters will be updated
    macro_name.setText('ascan')
    arg0.setText('gcdmot1')
    arg1.setText('1')
    arg2.setText('10')
    arg3.setText('5')
    arg4.setText('0.1')

    w.show()
    sys.exit(app.exec_())



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""
globals.py: 
"""

ORGANIZATION_NAME = "Taurus"

MACROEXECUTOR_APPLICATION_NAME = "macroexecutor"
SEQUENCER_APPLICATION_NAME = "sequencer"

SEQUENCES_DIRNAME = "Sequences"
SEQUENCE_DEFAULT_FILENAME = "Untitled.xml"

FAVOURITES_FILENAME = "favourite_macros.xml"

MACROS_HISTORY_FILENAME = ".macros_history"
MACROS_HISTORY_LENGHT = 5

TITLE_MACROSERVER_DIALOG = "Choose MacroServer"
TITLE_OUTPUT_DOCK_WIDGET = "Output"
TITLE_SPOCK_DOCK_WIDGET = "Spock"
TITLE_DEBUG_DOCK_WIDGET = "Debug"
TITLE_RESULT_DOCK_WIDGET = "Result"
TITLE_DESCRIPTION_DOCK_WIDGET = "Macro description"
TITLE_SEQUENCE_EDITOR_DOCK_WIDGET = "Sequence editor"
TITLE_MACRO_EXECUTOR_DOCK_WIDGET = "Macro executor"

MENU_FILE = "File"
MENU_FILE_CONFIGURATION = "Configuration"
MENU_FILE_NEW = "New"
MENU_FILE_OPEN = "Open..."
MENU_FILE_SAVE = "Save"
MENU_FILE_SAVEAS = "Save as..."
MENU_FILE_QUIT = "Quit"

MENU_CONTROL = "Control"
MENU_CONTROL_PLAY_SEQUENCE = "Play sequence"
MENU_CONTROL_PLAY_MACRO = "Play macro"
MENU_CONTROL_PAUSE = "Pause"
MENU_CONTROL_STOP = "Stop"

MENU_FAVOURITES = "Add to favourites"

MENU_VIEW = "View"

MENU_SETTINGS = "Settings"
MENU_SETTINGS_ONLYSCANS = "Only scan macros"

TIP_MENU_FILE_CONFIGURATION = "Configuration"
TIP_MENU_FILE_QUIT = "Quit application"
TIP_MENU_FILE_NEW = "New sequence"
TIP_MENU_FILE_OPEN = "Open sequence..."
TIP_MENU_FILE_SAVE = "Save sequence"
TIP_MENU_FILE_SAVEAS = "Save sequence as..."

TIP_MENU_CONTROL_PLAY_SEQUENCE = "Play sequence"
TIP_MENU_CONTROL_PLAY_MACRO = "Play macro"
TIP_MENU_CONTROL_PAUSE = "Pause"
TIP_MENU_CONTROL_STOP = "Stop"

TIP_MENU_FAVOURITES = "Add to favourites"


SHORTCUT_MENU_FILE_CONFIGURATION = "F2"
SHORTCUT_MENU_FILE_NEW = "Ctrl+N"
SHORTCUT_MENU_FILE_OPEN = "Ctrl+O"
SHORTCUT_MENU_FILE_SAVE = "Ctrl+S"
SHORTCUT_MENU_FILE_SAVEAS = "Ctrl+Shift+S"
SHORTCUT_MENU_FILE_QUIT = "Ctrl+Q"

SCAN_MACROS_FILTER = "scan"
ALLOWS_HOOKS_FILTER = 'allowsHooks'


PARAM_REPEAT = "ParamRepeat"

PARAM_OBJECT = "Object"
PARAM_MOTOR = "Motor"
PARAM_PSEUDOMOTOR = "PseudoMotor"
PARAM_MOVEABLE = "Moveable"
PARAM_CONTROLLER_CLASS = "ControllerClass"
PARAM_CONTROLLER = "Controller"
PARAM_MOTOR_PARAM = "MotorParam"
PARAM_MEASUREMENT_GROUP = "MeasurementGroup"
PARAM_IO_REGISTER = "IORegister"
PARAM_COMMUNICATION_CHANNEL = "ComChannel"
PARAM_EXPERIMENTAL_CHANNEL = "ExpChannel"
PARAM_MACRO_CODE = "MacroCode"
PARAM_MACRO_CLASS = "MacroClass"
PARAM_MACRO_FUNCTION = "MacroFunction"
PARAM_MACRO_LIBRARY = "MacroLibrary"


PARAM_STRING = "String"

PARAM_FILENAME = "FileName"

PARAM_FILENAME_2 = "Filename"

PARAM_INTEGER = "Integer"

PARAM_FLOAT = "Float"

PARAM_FILEDIALOG = "File"

PARAM_USERNAME = "User"

PARAM_BOOLEAN = "Boolean"

MAX_REPEATS = "max"
MIN_REPEATS = "min"

EDITOR_COMBOBOX_PARAMS = [PARAM_OBJECT,
                          PARAM_MOTOR,
                          PARAM_PSEUDOMOTOR,
                          PARAM_MOVEABLE,
                          PARAM_CONTROLLER_CLASS,
                          PARAM_CONTROLLER,
                          PARAM_MOTOR_PARAM,
                          PARAM_MEASUREMENT_GROUP,
                          PARAM_IO_REGISTER,
                          PARAM_COMMUNICATION_CHANNEL,
                          PARAM_EXPERIMENTAL_CHANNEL,
                          PARAM_MACRO_CODE,
                          PARAM_MACRO_CLASS,
                          PARAM_MACRO_FUNCTION,
                          PARAM_MACRO_LIBRARY]

EDITOR_LINEEDIT_PARAMS = [PARAM_STRING,
                          PARAM_FILENAME,
                          PARAM_FILENAME_2,
                          "Env"]

EDITOR_SPINBOX_PARAMS = [PARAM_INTEGER]

EDITOR_DOUBLESPINBOX_PARAMS = [PARAM_FLOAT]

EDITOR_FILEDIALOG_PARAMS = [PARAM_FILEDIALOG]

EDITOR_BOOLEAN_PARAMS = [PARAM_BOOLEAN]

EDITOR_NONEDITABLE_PARAMS = [PARAM_USERNAME]

TAG_MACRO = 'macro'
TAG_PARAM = 'param'
TAG_PARAMREPEAT = 'paramrepeat'
TAG_REPEAT = 'repeat'
TAG_PARAMS = 'params'
TAG_SEQUENCE = 'sequence'
TAG_ALLOWED_HOOK = 'allowedHook'
TAG_HOOK = 'hookPlaces'

ATTRIBUTE_NAME = 'name'
ATTRIBUTE_VALUE = 'value'
ATTRIBUTE_INDEX = 'nr'
ATTRIBUTE_DESCRIPTION = 'description'
ATTRIBUTE_DEFVALUE = 'defvalue'
ATTRIBUTE_TYPE = 'type'
ATTRIBUTE_ALLOWEDHOOKS = 'allowedHooks'
ATTRIBUTE_HASPARAMS = 'hasParams'
ATTRIBUTE_MIN = 'min'
ATTRIBUTE_MAX = 'max'



#!/usr/bin/env python
from sardana.sardanavalue import SardanaValue

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module is part of the Python Pool library. It defines the
PoolPseudoMotor class"""

__all__ = ["PoolPseudoMotor", "PoolPseudoMotorFrontend"]

__docformat__ = 'restructuredtext'

import sys
import time
import collections

from sardana import State, ElementType, TYPE_PHYSICAL_ELEMENTS
from sardana.sardanaattribute import SardanaAttribute
from sardana.sardanaexception import SardanaException

from sardana.pool.poolbaseelement import PoolBaseElement
from sardana.pool.poolelement import PoolElement
from sardana.pool.poolbasegroup import PoolBaseGroup
from sardana.pool.poolmotion import PoolMotion
from sardana.pool.poolexception import PoolException


class Position(SardanaAttribute):

    def __init__(self, *args, **kwargs):
        self._exc_info = None
        super(Position, self).__init__(*args, **kwargs)

        # 130226: We found a bug https://sourceforge.net/p/sardana/tickets/2/ that makes the Pool segfault with some pseudomotor configuration:
        # It can be reproduced by:
        # 4 physical motors: m1 m2 m3 m4
        # 2 slits: s1g,s1o = f(m1,m2) and s2g,s2o = f(m3,m4)
        # The pool will not be able to start if we create a third slit with s3g,s3o = f(s1g, s2g)
        # self.obj.get_physical_position_attribute_iterator() raises a KeyError exception
        # so we will flag the Position object as no_listeners for later configuration
        # We should still investigate the root of the problem when ordering the creation of elements
        self._listeners_configured = False
        try:
            for position_attr in self.obj.get_physical_position_attribute_iterator():
                position_attr.add_listener(self.on_change)
            self._listeners_configured = True
        except KeyError:
            pass

    def _in_error(self):
        for position_attr in self.obj.get_physical_position_attribute_iterator():
            if position_attr.error:
                return True
        return self._exc_info != None

    def _has_value(self):
        for position_attr in self.obj.get_physical_position_attribute_iterator():
            if not position_attr.has_value():
                return False
        return True

    def _has_write_value(self):
        for position_attr in self.obj.get_physical_position_attribute_iterator():
            if not position_attr.has_write_value():
                return False
        return True

    def _get_value(self):
        return self.calc_pseudo().value

    def _set_value(self, value, exc_info=None, timestamp=None, propagate=1):
        raise Exception("Cannot set position value for %s" % self.obj.name)

    def _get_write_value(self):
        w_positions = self.get_physical_write_positions()
        return self.calc_pseudo(physical_positions=w_positions).value

    def _set_write_value(self, w_value, timestamp=None, propagate=1):
        raise Exception("Cannot set position write value for %s" % self.obj.name)

    def _get_exc_info(self):
        exc_info = self._exc_info
        if exc_info is None:
            for position_attr in self.obj.get_physical_position_attribute_iterator():
                if position_attr.error:
                    return position_attr.get_exc_info()
        return exc_info

    def _get_timestamp(self):
        timestamps = [ pos_attr.timestamp for pos_attr in self.obj.get_physical_position_attribute_iterator() ]
        if not len(timestamps):
            timestamps = self._local_timestamp,
        return max(timestamps)

    def get_physical_write_positions(self):
        ret = []
        for pos_attr in self.obj.get_physical_position_attribute_iterator():
            if pos_attr.has_write_value():
                value = pos_attr.w_value
            else:
                if not pos_attr.has_value():
                    # if underlying moveable doesn't have position yet, it is
                    # because of a cold start
                    pos_attr.update(propagate=0)
                if pos_attr.in_error():
                    raise PoolException("Cannot get '%' position" % pos_attr.obj.name,
                                        exc_info=pos_attr.exc_info)
                value = pos_attr.value
            ret.append(value)
        return ret

    def get_physical_positions(self):
        ret = []
        for pos_attr in self.obj.get_physical_position_attribute_iterator():
            # if underlying moveable doesn't have position yet, it is because
            # of a cold start
            if not pos_attr.has_value():
                pos_attr.update(propagate=0)
            if pos_attr.in_error():
                raise PoolException("Cannot get '%' position" % pos_attr.obj.name,
                                    exc_info=pos_attr.exc_info)
            ret.append(pos_attr.value)
        return ret

    def calc_pseudo(self, physical_positions=None):
        try:
            obj = self.obj
            if physical_positions is None:
                physical_positions = self.get_physical_positions()
            else:
                l_p, l_u = len(physical_positions), len(obj.get_user_elements())
                if l_p != l_u:
                    raise IndexError("CalcPseudo(%s): must give %d physical " \
                                     "positions (you gave %d)" % (obj.name, l_u, l_p))
            result = obj.controller.calc_pseudo(obj.axis, physical_positions, None)
        except SardanaException as se:
            result = SardanaValue(exc_info=se.exc_info)
        except:
            result = SardanaValue(exc_info=sys.exc_info())
        return result

    def calc_all_pseudo(self, physical_positions=None):
        try:
            obj = self.obj
            if physical_positions is None:
                physical_positions = self.get_physical_positions()
            else:
                l_p, l_u = len(physical_positions), len(obj.get_user_elements())
                if l_p != l_u:
                    raise IndexError("CalcAllPseudo():: must give %d physical " \
                                     "positions (you gave %d)" % (l_u, l_p))
            result = obj.controller.calc_all_pseudo(physical_positions, None)
        except SardanaException as se:
            result = SardanaValue(exc_info=se.exc_info)
        except:
            result = SardanaValue(exc_info=sys.exc_info())
        return result

    def calc_physical(self, new_position):
        try:
            obj = self.obj
            curr_physical_positions = self.get_physical_positions()
            if isinstance(new_position, collections.Sequence):
                new_positions = new_position
            else:
                positions = obj.get_siblings_positions()
                positions[obj] = new_position
                new_positions = len(positions) * [None]
                for pseudo, position in positions.items():
                    new_positions[pseudo.axis - 1] = position

            result = obj.controller.calc_all_physical(new_positions,
                                                      curr_physical_positions)
        except SardanaException as se:
            result = SardanaValue(exc_info=se.exc_info)
        except:
            result = SardanaValue(exc_info=sys.exc_info())
        return result

    def on_change(self, evt_src, evt_type, evt_value):
        self.fire_read_event(propagate=evt_type.priority)

    def update(self, cache=True, propagate=1):
        if cache:
            for phy_elem_pos in self.obj.get_low_level_physical_position_attribute_iterator():
                if not phy_elem_pos.has_value():
                    cache = False
                    break
        if not cache:
            dial_position_values = self.obj.motion.read_dial_position(serial=True)
            if not len(dial_position_values):
                self._local_timestamp = time.time()
            for motion_obj, position_value in dial_position_values.items():
                motion_obj.put_dial_position(position_value, propagate=propagate)


class PoolPseudoMotor(PoolBaseGroup, PoolElement):
    """A class representing a Pseudo Motor in the Sardana Device Pool"""

    def __init__(self, **kwargs):
        self._siblings = None
        self._in_start_move = False
        self._drift_correction = kwargs.pop('drift_correction', None)
        user_elements = kwargs.pop('user_elements')
        kwargs['elem_type'] = ElementType.PseudoMotor
        PoolElement.__init__(self, **kwargs)
        PoolBaseGroup.__init__(self, user_elements=user_elements,
                               pool=kwargs['pool'])
        self._position = Position(self, listeners=self.on_change)

    # --------------------------------------------------------------------------
    # Event forwarding
    # --------------------------------------------------------------------------

    def on_change(self, evt_src, evt_type, evt_value):
        # forward all events coming from attributes to the listeners
        self.fire_event(evt_type, evt_value)

    def serialize(self, *args, **kwargs):
        kwargs = PoolElement.serialize(self, *args, **kwargs)
        elements = [ elem.name for elem in self.get_user_elements() ]
        physical_elements = []
        for elem_list in self.get_physical_elements().values():
            for elem in elem_list:
                physical_elements.append(elem.name)
        cl_name = self.__class__.__name__
        cl_name = cl_name[4:]
        kwargs['elements'] = elements
        kwargs['physical_elements'] = physical_elements
        return kwargs

    def _create_action_cache(self):
        motion_name = "%s.Motion" % self._name
        return PoolMotion(self, motion_name)

    def set_drift_correction(self, drift_correction):
        self._drift_correction = drift_correction

    def get_drift_correction(self):
        dc = self._drift_correction
        if dc is None:
            dc = self.manager.drift_correction
        return dc

    drift_correction = property(get_drift_correction,
                                set_drift_correction,
                                doc="drift correction")

    def get_action_cache(self):
        return self._get_action_cache()

    def set_action_cache(self, action_cache):
        self._set_action_cache(action_cache)

    def get_siblings(self):
        if self._siblings is None:
            self._siblings = siblings = set()
            for axis, sibling in self.controller.get_element_axis().items():
                if axis == self.axis:
                    continue
                siblings.add(sibling)
        return self._siblings

    siblings = property(fget=get_siblings,
                        doc="the siblings for this pseudo motor")

    def on_element_changed(self, evt_src, evt_type, evt_value):
        name = evt_type.name.lower()
        # always calculate state.
        status_info = self._calculate_states()
        state, status = self.calculate_state_info(status_info=status_info)
        state_propagate = 0
        status_propagate = 0
        if name == 'state':
            state_propagate = evt_type.priority
        elif name == 'status':
            status_propagate = evt_type.priority
        self.set_state(state, propagate=state_propagate)
        self.set_status(status, propagate=status_propagate)

    def add_user_element(self, element, index=None):
        elem_type = element.get_type()
        if elem_type == ElementType.Motor:
            pass
        elif elem_type == ElementType.PseudoMotor:
            #TODO: make this happen
            pass
        else:
            raise Exception("element %s is not a motor" % element.name)

        PoolBaseGroup.add_user_element(self, element, index=index)

    # ------------------------------------------------------------------------
    # position
    # ------------------------------------------------------------------------

    def calc_pseudo(self, physical_positions=None):
        return self.get_position_attribute().calc_pseudo(physical_positions=physical_positions)

    def calc_physical(self, new_position):
        return self.get_position_attribute().calc_physical(new_position)

    def calc_all_pseudo(self, physical_positions=None):
        return self.get_position_attribute().calc_all_pseudo(physical_positions=physical_positions)

    def get_position_attribute(self):
        return self._position

    def get_low_level_physical_position_attribute_iterator(self):
        return self.get_physical_elements_attribute_iterator()

    def get_physical_position_attribute_iterator(self):
        return self.get_user_elements_attribute_iterator()

    def get_physical_positions_attribute_sequence(self):
        return self.get_user_elements_attribute_sequence()

    def get_physical_positions_attribute_map(self):
        return self.get_user_elements_attribute_map()

    def get_physical_positions(self, cache=True, propagate=1):
        """Get positions for underlying elements.

        :param cache:
            if ``True`` (default) return value in cache, otherwise read value
            from hardware
        :type cache:
            bool
        :param propagate:
            0 for not propagating, 1 to propagate, 2 propagate with priority
        :type propagate:
            int
        :return:
            the physical positions
        :rtype:
            dict <PoolElement, :class:`~sardana.sardanaattribute.SardanaAttribute` >"""
        self._position.update(cache=cache, propagate=propagate)
        return self.get_physical_positions_attribute_map()

    def get_siblings_positions(self, use=None, write_pos=True):
        """Get the last positions for all siblings.
        If write_pos is True and a sibling has already been moved before,
        it's last write position is used. Otherwise its read position is used
        instead.
        
        :param use: the already calculated positions. If a sibling is in this
                    dictionary, the position stored here is used instead
        :type use: dict <PoolElement, :class:`~sardana.sardanavalue.SardanaValue` >
        :param write_pos: determines if should try to use the last set point
                          [default: True]
        :type write_pos: bool
        :return: a dictionary with siblings write positions
        :rtype:
            dict <PoolElement, position(float?) >"""
        positions = {}
        for sibling in self.siblings:
            pos_attr = sibling.get_position(propagate=0)
            if use and sibling in use:
                pos = use[sibling]
            elif pos_attr.has_write_value() and write_pos:
                pos = pos_attr.w_value
            else:
                if pos_attr.in_error():
                    raise PoolException("Cannot get '%s' position" % sibling.name,
                                        exc_info=pos_attr.exc_info)
                pos_value = pos_attr.calc_pseudo()
                if pos_value.error:
                    raise PoolException("Cannot get '%s' position" % sibling.name,
                                        exc_info=pos_value.exc_info)
                pos = pos_value.value
            positions[sibling] = pos
        return positions

    def get_position(self, cache=True, propagate=1):
        """Returns the user position.

        :param cache:
            if ``True`` (default) return value in cache, otherwise read value
            from hardware
        :type cache:
            bool
        :param propagate:
            0 for not propagating, 1 to propagate, 2 propagate with priority
        :type propagate:
            int
        :return:
            the user position
        :rtype:
            :class:`~sardana.sardanaattribute.SardanaAttribute`"""
        position_attr = self._position
        position_attr.update(cache=cache, propagate=propagate)
        return position_attr

    def set_position(self, position):
        """Moves the motor to the specified user position

        :param position:
            the user position to move to
        :type position:
            :class:`~numbers.Number`"""
        self.start_move(position)

    def set_write_position(self, w_position, timestamp=None, propagate=1):
        """Sets a new write value for the user position.

        :param w_position:
            the new write value for user position
        :type w_position:
            :class:`~numbers.Number`
        :param propagate:
            0 for not propagating, 1 to propagate, 2 propagate with priority
        :type propagate:
            int"""
        # SardanaAttribute position will raise an exception so we let it do it
        self._position.set_write_value(w_position, timestamp=timestamp,
                                       propagate=propagate)

    position = property(get_position, set_position, doc="pseudo motor position")

    # ------------------------------------------------------------------------
    # state information
    # ------------------------------------------------------------------------

    _STD_STATUS = "{name} is {state}\n{ctrl_status}"
    def calculate_state_info(self, status_info=None):

        # Refer to Position.__init__ method for an explanation on this 'hack'
        if not self._position._listeners_configured:
            for position_attr in self.get_physical_position_attribute_iterator():
                position_attr.add_listener(self._position.on_change)
            self._position._listeners_configured = True

        if status_info is None:
            status_info = self._state, self._status
        state, status = status_info
        if state == State.On:
            state_str = "Stopped"
        else:
            state_str = "in " + State[state]
        new_status = self._STD_STATUS.format(name=self.name, state=state_str,
                                             ctrl_status=status)
        return status_info[0], new_status

    def read_state_info(self, state_info=None):
        if state_info is None:
            state_info = {}
            action_cache = self.get_action_cache()
            ctrl_state_infos = action_cache.read_state_info(serial=True)
            for motion_obj, ctrl_state_info in ctrl_state_infos.items():
                state_info[motion_obj] = motion_state_info = \
                    motion_obj._from_ctrl_state_info(ctrl_state_info)
                motion_obj.put_state_info(motion_state_info)
        for user_element in self.get_user_elements():
            if user_element.get_type() not in TYPE_PHYSICAL_ELEMENTS:
                motion_state_info = user_element._calculate_states()
                user_element.put_state_info(motion_state_info)

        ret = self._calculate_states()
        return ret

    # --------------------------------------------------------------------------
    # default acquisition channel
    # --------------------------------------------------------------------------

    def get_default_attribute(self):
        return self.get_position_attribute()

    # ------------------------------------------------------------------------
    # motion
    # ------------------------------------------------------------------------

    def get_motion(self):
        return self.get_action_cache()

    motion = property(get_motion, doc="motion object")

    # ------------------------------------------------------------------------
    # motion calculation
    # ------------------------------------------------------------------------

    def calculate_motion(self, new_position, items=None, calculated=None):
        # if items already contains the positions for this pseudo motor
        # underlying motors it means the motion has already been calculated
        # by a sibling
        if items is not None and len(items):
            physical_elements = self.get_physical_elements_set()
            s_items = set(items)
            if s_items == physical_elements:
                if calculated is not None and self in calculated:
                    return

        user_elements = self.get_user_elements()
        positions = self.get_siblings_positions(use=calculated,
                                                write_pos=self.drift_correction)
        positions[self] = new_position
        pseudo_positions = len(positions) * [None]
        for pseudo, position in positions.items():
            pseudo_positions[pseudo.axis - 1] = position
        curr_physical_positions = self._position.get_physical_positions()
        physical_positions = self.controller.calc_all_physical(pseudo_positions,
                                                               curr_physical_positions)
        if physical_positions.error:
            raise PoolException("Cannot calculate motion: "
                                "calc_all_physical raises exception",
                                exc_info=physical_positions.exc_info)
        else:
            if physical_positions.value is None:
                raise PoolException("Cannot calculate motion: "
                                    "calc_all_physical returns None")

        if items is None:
            items = {}
        for new_position, element in zip(physical_positions.value, user_elements):
            if new_position is None:
                raise PoolException("Cannot calculate motion: %s reports "
                                     "position to be None" % element.name)
            element.calculate_motion(new_position, items=items,
                                     calculated=calculated)
        return items

    def start_move(self, new_position):
        self._in_start_move = True
        try:
            return self._start_move(new_position)
        finally:
            self._in_start_move = False

    def _start_move(self, new_position):
        self._aborted = False
        self._stopped = False
        items = self.calculate_motion(new_position)
        timestamp = time.time()
        for item, position_info in items.items():
            item.set_write_position(position_info[0], timestamp=timestamp,
                                    propagate=1)
        if not self._simulation_mode:
            self.motion.run(items=items)

    # ------------------------------------------------------------------------
    # stop
    # ------------------------------------------------------------------------

    def stop(self):
        #surpass the PoolElement.stop because it doesn't do what we want
        PoolBaseElement.stop(self)
        PoolBaseGroup.stop(self)

    # ------------------------------------------------------------------------
    # abort
    # ------------------------------------------------------------------------

    def abort(self):
        #surpass the PoolElement.abort because it doesn't do what we want
        PoolBaseElement.abort(self)
        PoolBaseGroup.abort(self)

    # ------------------------------------------------------------------------
    # involved in an operation
    # ------------------------------------------------------------------------

    def get_operation(self):
        return PoolBaseGroup.get_operation(self)



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This file contains the pool monitor class"""

__all__ = ["PoolMonitor"]

__docformat__ = 'restructuredtext'

import time
import threading

from taurus.core.util.log import Logger

from sardana import ElementType, TYPE_PSEUDO_ELEMENTS

from sardana.pool.poolobject import PoolObject


class PoolMonitor(Logger, threading.Thread):

    MIN_THREADS = 1
    MAX_THREADS = 10

    def __init__(self, pool, name='PoolMonitor', period=5.0, min_sleep=1.0,
                 auto_start=True):
        Logger.__init__(self, name)
        threading.Thread.__init__(self, name=name)
        self.daemon = True
        self._period = period
        self._min_sleep = min_sleep
        self._pool = pool
        self._stop = False
        self._pause = threading.Event()
        self._thread_pool = None
        self._ctrl_ids = []
        self._elem_ids = []
        pool.add_listener(self.on_pool_changed)
        if not auto_start:
            self.pause()
        self.start()

    def on_pool_changed(self, evt_src, evt_type, evt_value):
        evt_name = evt_type.name.lower()
        if "created" in evt_name or "deleted" in evt_name:
            pool = self._pool
            pool_ctrls = pool.get_elements_by_type(ElementType.Controller)
            pool_ctrls.sort(key=PoolObject.get_id)
            ctrl_ids = []
            elem_ids = []
            for pool_ctrl in pool_ctrls:
                if not pool_ctrl.is_online():
                    continue
                types = set(pool_ctrl.get_ctrl_types())
                if types.isdisjoint(TYPE_PSEUDO_ELEMENTS):
                    ctrl_ids.append(pool_ctrl.id)
                    elem_ids.extend(pool_ctrl.get_element_ids().keys())
            elem_ids.sort()
            self._elem_ids = elem_ids
            self._ctrl_ids = ctrl_ids

    def update_state_info(self):
        """Update state information of every element."""

        pool = self._pool
        elems, ctrls, ctrl_items = [], [], {}
        try:
            blocked_ctrls = set()
            for elem_id in self._elem_ids:
                elem = pool.get_element_by_id(elem_id)
                ctrl = elem.controller
                if elem.is_in_operation():
                    blocked_ctrls.add(ctrl)
                    continue
                if ctrl in blocked_ctrls:
                    continue
                ret = elem.lock(blocking=False)
                if ret:
                    elems.append(elem)
                    ctrl_elems = ctrl_items.get(ctrl)
                    if ctrl_elems is None:
                        ctrl_items[ctrl] = ctrl_elems = []
                    ctrl_elems.append(elem)
                else:
                    blocked_ctrls.add(ctrl)

            for ctrl, ctrl_elems in ctrl_items.items():
                ret = ctrl.lock(blocking=False)
                if ret:
                    ctrls.append(ctrl)
                else:
                    for elem in reversed(ctrl_elems):
                        elem.unlock()
                        elems.remove(elem)

            self._update_state_info_serial(ctrl_items)
        finally:
            for ctrl in reversed(ctrls):
                ctrl.unlock()
            for elem in reversed(elems):
                elem.unlock()

    def _update_state_info_serial(self, pool_ctrls):
        for pool_ctrl, elems in pool_ctrls.items():
            self._update_ctrl_state_info(pool_ctrl, elems)

    def _update_ctrl_state_info(self, pool_ctrl, elems):
        axes = [elem.axis for elem in elems]
        state_infos, exc_info = pool_ctrl.raw_read_axis_states(axes)
        if len(exc_info):
            self.info("STATE ERROR %s", exc_info)
        for elem, state_info in state_infos.items():
            state_info = elem._from_ctrl_state_info(state_info)
            elem.set_state_info(state_info)

    def stop(self):
        self.resume()
        self._stop = True

    def pause(self):
        self._pause.clear()

    def resume(self):
        self._pause.set()

    def monitor(self):
        ret = self.update_state_info()

    def run(self):
        nap_time = period = self._period
        i, startup = 0, time.time()
        while True:
            if self._stop:
                break
            time.sleep(nap_time)
            self._pause.wait()
            self.monitor()
            finish = time.time()
            nap_time = -1
            while nap_time < 0:
                i += 1
                nap_time = (startup + i * self._period) - finish



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""Tests for wm macros"""

from taurus.external import unittest
from sardana.macroserver.macros.test import (RunMacroTestCase, testRun,
                                             SarDemoEnv)

try:
    _MOTORS = SarDemoEnv().getMotors()
    _m1, _m2 = _MOTORS[:2]
except RuntimeError:
    import taurus
    from sardana import sardanacustomsettings
    door_name = getattr(sardanacustomsettings, 'UNITTEST_DOOR_NAME',
                        'UNDEFINED')
    taurus.warning("The door %s is not running. " % (door_name) +
                   "Ignore this message if you are building the documentation.")
    _m1 = _m2 = 'motor_not_defined'
except Exception, e:
    import taurus
    taurus.debug(e)
    taurus.warning("It was not possible to retrieve the motor names. " +
                 "Ignore this message if you are building the documentation.")
    _m1 = _m2 = 'motor_not_defined'


class WBase(RunMacroTestCase):

    """Base class for testing macros used to read position.
    """

    def macro_runs(self, **kw):
        """Testing the execution of the 'wm' macro and verify that the log
        'output' exists.
        """
        RunMacroTestCase.macro_runs(self, **kw)
        self.logOutput = self.macro_executor.getLog("output")
        msg = "wm macro did not return any data."
        self.assertTrue(len(self.logOutput) > 0, msg)


@testRun(macro_params=[_m1], wait_timeout=5.0)
class WmTest(WBase, unittest.TestCase):

    """Test of wm macro. It verifies that the macro 'wm' can be executed.
    It inherits from WmBase and from unittest.TestCase.
    It tests the execution of the 'wm' macro and verifies that the log 'output'
    exists.
    """
    macro_name = "wm"



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""test_parameter module documentation"""

from taurus.external import unittest
from sardana.spock import parameter


class ParamTestCase(unittest.TestCase):

    """
    Instantiate in different ways a Param object and verify that
    they are correct instances from the class Param.
    """

    def testInstanceCreation(self):
        """
        Instantiate in different ways a Param object.
        """
        spock_param = parameter.Param()
        self.assertIsInstance(spock_param, parameter.Param,
                              'Instantiation of an object Param without arguments ' +
                              'does not work \n')

        spock_param = parameter.Param(name='sardanaName')
        self.assertIsInstance(spock_param, parameter.Param,
                              'Instantiation of an object Param with argument name ' +
                              'does not work')

        spock_param = parameter.Param(name='sardanaName',
                                      desc='description_is_present')
        self.assertIsInstance(spock_param, parameter.Param,
                              'Instantiation of an object Param with arguments name ' +
                              'and description does not work')

        spock_param = parameter.Param(name='sardanaName',
                                      desc='description_is_present', type_name='integer')
        self.assertIsInstance(spock_param, parameter.Param,
                              'Instantiation of an object Param with arguments name, ' +
                              'description and type_name does not work')

        spock_param = parameter.Param(name='sardanaName',
                                      desc='description_is_present', type_name='integer', defvalue=7)
        self.assertIsInstance(spock_param, parameter.Param,
                              'Instantiation of an object Param with arguments name, ' +
                              'description, type_name and defvalue does not work')
